import java.util.*;
import java.io.*;
/**
 * BidSet.java
 *
 * A BidSet is a set of Bid objects.
 *
 * Attributes with a "DS" prefix no longer accumulate as MAX. See
 * makeOneBidFromBidSet().  DS must accumulate as sum so that two
 * trees have more DS than one; otherwise DS goes down relative to
 * cost and combinations have lower utility than the parts.
 * 
 * Created: Mon Jul 28 13:52:57 2003
 * 
 * An empty BidSet displays as the String "empty".
 *
 * @author <a href="mailto:terry@Terry-Harveys-Computer.local.">Terry Harvey</a>
 * @version 1.0
 */
public class BidSet extends HashSet {

    static int suffix = -1; //used to give unique names to sets
    String name;
    private String owner = "OwnerNotSet";

//the default min U before a bid is considered to have failed; used if
//there are no real bids for comparison. Use higher envelope of real alternative or this percentage.
    static double minEnvelopePercent = 99; 
    static boolean DEBUG = false;
    static boolean SHOW_CLUSTERS = false;
    static String em = "Error in BidSet.java: ";

    public BidSet() {
	name = getUniqueName();
    } // BidSet constructor
    
    public BidSet(String ownr){ 
	owner = ownr;
	name = getUniqueName(); 
    }
    
    public BidSet(String ownr, Collection c){
	super(c);
	owner = ownr;
	name = getUniqueName(); 
    }

    /**Standard HashSet clone() is shallow, but this is deep. **/
    public Object clone(){
	BidSet result = new BidSet();//(BidSet)super.clone();
	result.owner = owner;
	result.name = getUniqueName(); 
	Iterator i = this.iterator();
	while (i.hasNext())
	    result.add( ((Bid)i.next()).clone() );
	return result;
    }

    /** this didn't help memory performance - why? **/
    public Object shallowClone(){
	BidSet result = (BidSet)super.clone();
	result.owner = owner;
	result.name = getUniqueName();
	return result;
    }

    public String getOwner(){ return owner; }
    public void setOwner(String ownr){ owner = ownr; }

    static String getUniqueName(){
	suffix += 1;
	return "CS" + suffix;
    }

    /**
     * This function should not be necessary. However, as of 1.4 I was
     * getting inexplicable behavior from both hashset.remove and
     * iterator.remove. Even when asked to remove an item ("mine", see
     * below) that is known to be inside the set, both removes
     * failed. The bad behavior was in PL_Count.manageFailure().
     **/
    //public boolean remove(Object o){
    //	//System.out.println("my remove");
    //	if (o instanceof Bid){
    //	    Bid mine = null;
    //	    Vector v = new Vector(this.size());
    //	    Iterator i = this.iterator();
    //	    while (i.hasNext()){
    //		mine = (Bid)i.next();
    //		if (mine.equals((Bid)o)){
    //		    if (DEBUG) System.out.println("not adding " + mine);
    //		}
    //		else
    //		    v.add(mine);
    //	    }
    //	    //this.clear();
    //	    //this.addAll(v);
    //	    //return true;
    //	    if (mine != null){
    //	    	System.out.println("calling super remove on " + mine);
    //		return super.remove(mine);
    //	    }
    //	    else return false;
    //	}
    //	return super.remove(o);
    //}

    /** 
     * Remove all bids that do not meet constraints.
     */
    public void filterUsingConstraints(ConstraintSet cs){
	Iterator i = this.iterator();
	Bid b;
	while (i.hasNext()){
	    b = (Bid)i.next();
	    if (! b.isWithinConstraints(cs)){
		i.remove();
	    }
	}
    }

    public void setEachBidUtility(BidUtility bu){

	Vector v = new Vector(this);
	//v.addAll(this);
	for (int i=0; i<v.size(); i++)
	    ((Bid)v.elementAt(i)).getUtility(bu);
    }
    /**
     * Use carefully.
     *
     * Used by PL_Update when reading generically named local bids for multiple clone agents
     **/
    public void setEachBidOwner(String newOwner){

	Vector v = new Vector(this);
	//v.addAll(this);
	for (int i=0; i<v.size(); i++)
	    ((Bid)v.elementAt(i)).setBidder(newOwner);
    }

    /** Don't do this; make a single bid from set, then use its utility.
     * old: Sum the utilities of all bids in set.
     **/
    //public double addBidUtilities(BidUtility uFcn){
    //	double value, utility = 0;
    //	Iterator bids = this.iterator();
    //	while (bids.hasNext()){
    //	    value = ((Bid)bids.next()).getUtility(uFcn).doubleValue();
    //	    utility += value; 
    //	}
    //	return utility;
    //}

    public Bid getHighestUtilityBid(BidUtility bu){
	if (this.isEmpty()) return null;
	Vector v = new Vector(this);
	Bid highest = (Bid)v.elementAt(0);
	double highestU = highest.getUtility(bu).doubleValue();
	double nextU;
	for (int i=1; i<v.size(); i++){
	    nextU = ((Bid)v.elementAt(i)).getUtility(bu).doubleValue();
	    if (highestU < nextU){
		highest = (Bid)v.elementAt(i);
		highestU = nextU;
	    }
	}
	if (DEBUG) 
	    System.out.println(" getHighestUtilityBid returning: " + highest);
	return highest;
    }

    //takes a vector of Bids
    private Bid getHighestUtilityBid(BidUtility bu, Vector v){
	if (v.isEmpty()) return null;
	Bid highest = (Bid)v.elementAt(0);
	double highestU = highest.getUtility(bu).doubleValue();
	double nextU;
	for (int i=1; i<v.size(); i++){
	    nextU = ((Bid)v.elementAt(i)).getUtility(bu).doubleValue();
	    if (highestU < nextU){
		highest = (Bid)v.elementAt(i);
		highestU = nextU;
	    }
	}
	if (DEBUG) 
	    System.out.println(" getHighestUtilityV returning: " + highest);
	return highest;
    }

    /* NOT SORTED */ /* rewrite with selection sort? too many calls to Bid.getUtility*/
    public BidSet getNHighestUtility(int n, BidUtility bu){
	if (this.size() <= n)
	    return (BidSet)this.clone();

	BidSet bs = new BidSet();
	Vector v = new Vector(this);
	Bid next;

	for (int i=0; i<n; i++){
	    next = getHighestUtilityBid(bu, v);
	    bs.add(next);
	    v.remove(next);
	}
	if (DEBUG) System.out.println("getNHighestUtility returning unsorted: " + bs);
	return bs;
    }

    /* NOT SORTED */
    /**
     * Call Kruskal's to get n+1 clusters of bids most different in
     * parameter space. (The plus one is so that high bid cluster is not
     * one of the alts) Then from each cluster select highest U bid.
     *
     * COULD zero all bid values that have zero coeff in bu, so that
     * only values that matter to user are compared for distance
     * metric. However, the whole point is to generate alternatives
     * that will work under *other* utility functions, which may have
     * different coeffs. So don't do that.
     **/
    public BidSet getNAltUtility(int n, BidUtility bu, Bid hiBidToExclude){

	if (n <= 0)
	    return new BidSet(); //send back empty so union doesn't cause errors
	else if (this.size() <= n)
	    return (BidSet)this.clone();

	BidSet bsNew = new BidSet();
	Bid nextBid;

	//get Vector of n+1 clusters from Kruskal's
	//Vector vThis = new Vector(this);
	Kruskal k = new Kruskal(n + 1, this);
	Vector clusters = null;
	try{
	    clusters = k.getHashSetsOfObjects();//trees are HashSets of bids

	    if (SHOW_CLUSTERS){
		for (int i=0; i<clusters.size(); i++){
		    BidSet bs = new BidSet("test", (Collection)clusters.get(i) );
		    Vector v = bs.getNSortedHighestUtility(bs.size(), bu);
		    for (int j = 0; j<v.size(); j++)
			System.out.println(v.get(j));
		    System.out.println();
		    
		}
	    }

	    //extract high U bid from each cluster and insert in bsNew 
	    for (int i=0; i < clusters.size(); i++){
		nextBid = 
		    getHighestUtilityBid(bu, new Vector( (Collection)clusters.get(i) ));
		if ( ! nextBid.equals(hiBidToExclude) )
		    bsNew.add(nextBid);
	    }

	} catch (Exception e) {
	    System.out.println("***\n***\n***\n***\n***\n BidSet calling K.consolidate " + e);
	}

	//if (DEBUG) 
	    System.out.println("getNAltUtility returning unsorted: " + bsNew);
	return bsNew;
    }//getNAltUtility


    //FIX: make this private again, don't sort
    public Vector getNSortedHighestUtility(int n, BidUtility bu){
	
	Vector v = new Vector();
	v.addAll(this);
	for (int i=0; i<v.size(); i++)
	    ((Bid)v.elementAt(i)).getUtility(bu);
	UtilComp uc = new UtilComp(bu);
	Collections.sort(v, uc);
	
	if (v.size() < n)
	    n = v.size();
	List l = v.subList(0, n);
	Vector v1 = new Vector(n);
	v1.addAll(l);
	if (DEBUG) 
	    System.out.println("getNSortedHighestUtility returning: " + v1);
	return v1;
    }

    public Envelope getBestBidsEnvelope(BidUtility bu){

	if (this.isEmpty())
	    return new Envelope();
	
	Vector bestTwo = getNSortedHighestUtility(2, bu);
	Bid hi, low;
	if (DEBUG)System.out.println("inside getBestBidsEnvelope1, best: " + bestTwo);
	if (DEBUG)System.out.println("inside getBestBidsEnvelope1, size: " + bestTwo.size());
//	System.out.println("inside getBestBidsEnvelope1, sources: " + 
//			   ((Bid)bestTwo.elementAt(0)).getBidSetSource() + " " +
//these probably are not set      		   ((Bid)bestTwo.elementAt(1)).getBidSetSource());
	if (bestTwo.size() < 2){
	    if (DEBUG)System.out.println("inside getBestBidsEnvelope2, best: " + bestTwo.elementAt(0).toString());
	    
	    return makeDefaultEnvelope((Bid)bestTwo.firstElement(), bu);
	}
	else
	    return calcEnvelopeFromTwoBids((Bid)bestTwo.elementAt(0),
					   (Bid)bestTwo.elementAt(1),
					   bu);
    }//getBestBidsEnvelope(BidUtility bu)

    private Envelope calcEnvelopeFromTwoBids(Bid high, Bid low, BidUtility bu){
	Envelope result = new Envelope(high);
	double Uhigh = high.getUtility(bu).doubleValue();
	double Ulow = low.getUtility(bu).doubleValue();
	double percentMin = minEnvelopePercent/100;
	if (! (Uhigh >= Ulow))
	    System.out.println("WARNING: Envelope utilities are mixed up.");
	if (  Ulow > (percentMin * Uhigh) ){
	    System.out.println("Alternative U " + Ulow + 
			       " is too high; Using default envelope of " + 
			       percentMin + " percent");
	    result.put("utility_min", Double.toString(percentMin * Uhigh));
	}
	else{
	    result.put("utility_min", Double.toString(Ulow));
	}
	return result;
    }

    private Envelope makeDefaultEnvelope(Bid b, BidUtility bu){
	Envelope result = new Envelope(b);
	result.put("utility_min", Double.toString(minEnvelopePercent/100 * b.getUtility(bu).doubleValue()));
	return result;
    }

    public static BidSet makeSingleBidsFromBidSets(String bidder, Vector bsets){
	BidSet b = new BidSet(bidder);
	Iterator bidsets = bsets.iterator();
	while (bidsets.hasNext())
	    b.add(((BidSet)bidsets.next()).makeOneBidFromBidSet(bidder));
	return b;
    }

    /**
     * Make single bid to represent contents of this BidSet. 
     * FIX: each attribute should have its own accumulation fcn. 
     * partially fixed: "DS" prefix now accumulates as merged Strength. So does RV, time.
     * FIX: what if attribute list doesn't match?
     **/
    public Bid makeOneBidFromBidSet(String bidder){
	Bid result = new Bid(bidder);
	Bid current;
	Iterator attributes;
	String attName;
	double total = 0, next;
	Iterator bids = this.iterator();
	while (bids.hasNext()){
	    current = (Bid)bids.next();
	    attributes = current.getAttributes();
	    while (attributes.hasNext()){
		attName = (String)attributes.next();
		if (DEBUG) System.out.println("makeOneBidFromBidSet attribute name is: " + attName);
		
		if (result.containsKey(attName)){
		    
		    if (attName.startsWith("DS")){
			result.put(attName, Strength.merge((Strength)current.get(attName), (Strength)result.get(attName)));
		    }
		    else if (attName.equals("RV")){
			result.put(attName, RuleValue.merge((RuleValue)current.get(attName), (RuleValue)result.get(attName)));
		    }
		    else if (attName.startsWith("time")){
			total = Double.parseDouble((String)result.getValue(attName));
			next = Double.parseDouble((String)current.getValue(attName));
			result.put(attName, Double.toString(Math.max(total, next)));
			
		    }
		    else if (attName.equals("symbol")){
			result.put(attName, current.get(attName));
		    }
		    else if (attName.equals("topic")) continue;
		    else if (attName.equals("bidder")) continue;
		    else {		//	total = Math.max(total, Double.parseDouble((String)current.getValue(attName)));
			total = Double.parseDouble((String)result.getValue(attName));
			total += Double.parseDouble((String)current.getValue(attName));
			result.put(attName, Double.toString(total));
		    }
		}
		else result.put(attName, current.getValue(attName));
	    }//end while
	}
	result.setBidSetSource(this);
	return result;
    }

    /**
     * Returns a vector of BidSets.
     * Make all possible combinations using 0 or 1 element from each BidSet.  
     *;> (choose-1-combos '((1 2 3) (a b)))
     *;((1) (2) (3) (1 a) (1 b) (2 a) (2 b) (3 a) (3 b) (a) (b))
     **/
    public static Vector chooseOneCombos(Vector bidSets){
	return chooseOneCombos("OwnNotset", bidSets);}

    public static Vector chooseOneCombos(String owner, Vector bidSets){
	Vector result = new Vector(50);
	Vector copy = null, copy2 = null, newElements;
	BidSet cbs, temp;
	Bid bid;
	Iterator bids = bidSets.iterator();
	
	if (bids.hasNext()){	    //make a vector of bidsets, one bidset for each element of this first BidSet
	    cbs = (BidSet)bids.next();
	    Iterator i = cbs.iterator();
	    while (i.hasNext()){
		temp = new BidSet(owner);
		temp.add(i.next());
		result.add(temp);
	    }
	}
	if (DEBUG) System.out.println("first result: " + result);
			
	//Now process remaining BidSets
	
	while (bids.hasNext()){
	    copy = cloneVectorOfBidSets(result);
	    if (DEBUG) System.out.println("setting copy to: " + copy);
	    //make a vector of bidsets, one bidset for each element of current bidset
	    newElements = new Vector(5);
	    cbs = (BidSet)bids.next();
	    Iterator i = cbs.iterator();
	    while (i.hasNext()){
		bid = (Bid)i.next();
		temp = new BidSet();
		temp.add(bid);
		newElements.add(temp);
	    
		//for each member of current bidset, add to all previous results:
		copy2 = cloneVectorOfBidSets(copy);
		if (DEBUG) System.out.println("setting copy2 to: " + copy2);
		//add member to every member of copy, 
		addObjToEachBidSet(bid, copy2);
		if (DEBUG) System.out.println("       copy2 now: " + copy2);
		result.addAll(copy2);
		if (DEBUG) System.out.println("next result: " + result);

	    }
	    if (DEBUG) System.out.println("newElements: " + newElements);	    
	    result.addAll(newElements);
	    if (DEBUG) System.out.println("loop result: " + result);
	}
	if (DEBUG) System.out.println("final result: " + result);
	return result;
    }//chooseOneCombos(Vector bidSets)

    //helper fcn for chooseOneCombos
    private static void addObjToEachBidSet(Object o, Vector bidSets){
	Iterator i = bidSets.iterator();
	while (i.hasNext())
	    ((BidSet)i.next()).add(o);
    }

    //helper fcn for chooseOneCombos
    private static Vector cloneVectorOfBidSets(Vector v){
	Vector v1 = new Vector(v.size());
	BidSet b;
	Iterator i = v.iterator();
	while (i.hasNext()){
	    b = (BidSet)i.next();
	    v1.add(b.clone());
	}
	return v1;
    }

    //[{cost_exp=3, U=null}, {cost_exp=2, length_exp=2, U=null}, {cost_exp=1, length_exp=1, U=null}]
    static public BidSet readBidSetFromString(String s){
	String next;
	Bid b;
	BidSet bs = new BidSet();
	StringTokenizer st = new StringTokenizer(s, "[], {}=", true);
	while (st.hasMoreTokens()){
	    next = st.nextToken();
	    if (DEBUG) System.out.println(next);
	    if (next.equals("empty")) return null;
 	    if (next.equals("[") || next.equals("]") || next.equals(",") || next.equals(" "))
		continue;
 	    else if (next.equals("{")){ 
		b = Bid.readBid(st);
		bs.add(b);
	    }
	    else if (next.equals("owner")){
		next = st.nextToken();
		if (next.equals("="))
		    bs.setOwner(st.nextToken());
		else
		    System.out.println(em + "parsing BidSet, expecting =, got " + next);
	    }
	    else
		System.out.println(em + "parsing BidSet, expecting what?, got " + next);
	}
	return bs;
    }//readBidSetFromString(String s)
    
    //[{cost_exp=3, U=null}, {cost_exp=2, length_exp=2, U=null}, {cost_exp=1, length_exp=1, U=null} owner=Spam]
    ///already parsed first '['
    static public BidSet readBidSet(StringTokenizer st){
	String next;
	Bid b;
	BidSet bs = new BidSet();
	while (st.hasMoreTokens()){
	    next = st.nextToken();
	    if (DEBUG) System.out.println(next);
 	    if (next.equals("]"))
		return bs;
	    else if (next.equals(",") || next.equals(" "))
		continue;
 	    else if (next.equals("{")){ 
		b = Bid.readBid(st);
		bs.add(b);
	    }
	    else if (next.equals("owner")){
		next = st.nextToken();
		if (next.equals("="))
		    bs.setOwner(st.nextToken());
		else
		    System.out.println(em + "parsing BidSet, expecting =, got " + next);
	    }
	    else
		System.out.println(em + "parsing BidSet, expecting what?, got " + next);
	}
	return bs;
    }//readBidSetFromString(StringTokenizer st)
    
    /**
     * Read Bids, each on single line, from file, and return a BidSet.
     **/
    public static BidSet readBidsFromFile(String filename)
    {

	String next;
	Bid nextBid;
	BidSet bs = new BidSet();

	try {
	    BufferedReader in = new BufferedReader(new FileReader(filename));

	    next = in.readLine();
	    while (next != null){
		next = next.trim();
		if ( ! next.startsWith("//")){
		    nextBid = Bid.readBidFromString(next);
		    bs.add(nextBid);
		}
		next = in.readLine();
	    }

	} catch (IOException e){
	    e.printStackTrace();
	    System.out.println(e);
	}
	return bs;
    }

    //[[{cost_exp=2, length_exp=2, ...}, {cost_exp=1,...}], [{...}, {...}]]
    static public Vector readBidSetVector(String s){
	String next = null;
	BidSet bs;
	Vector v = new Vector(10);
	StringTokenizer st = new StringTokenizer(s, "[], {}=", true);

	if  (st.hasMoreTokens()){
	    next = st.nextToken();
	    while (next.equals(" ") && st.hasMoreTokens()) next = st.nextToken(); //trim
	}	

	if (next.equals("[")){
	    while (st.hasMoreTokens()){
		next = st.nextToken();
		//if (DEBUG) System.out.println(next);
		if ( next.equals("empty") || next.equals("]") || next.equals(",") || next.equals(" "))
		    continue;
		else if (next.equals("[")){ 
		    bs = BidSet.readBidSet(st);
		    v.add(bs);
		}
	    }
	    return v;
	}
	else 
	    System.out.println(em + "incorrect first token in readBidSetVector(String s): " + s);
	return null;
    }//readBidSetVector(String s)


    public String toString(){

	if (this.isEmpty()) return "empty";

	String s = super.toString();
	int size = s.length();
	s = s.substring(0, size - 1);
	return s + ", owner=" + owner + "]";
    }
    

    public static void main(String[] args){

	//Bid a = new Bid();
	//a.put("cost_exp", "1");	a.put("length_exp", "1");
	//
	//Bid b = new Bid();
	//b.put("cost_exp", "2");	b.put("length_exp", "2");
	//
	//Bid c = new Bid();
	//c.put("cost_exp", "3");	
	//
	//BidSet one = new BidSet("hakim");
	////	BidSet two = new BidSet("mary");
	////BidSet three = new BidSet("ethel");
	//one.add(a);
	//one.add(b);
	//one.add(c);
	////	Vector v1 = chooseOneCombos(v);
	//Bid x = one.makeOneBidFromBidSet("x");
	//System.out.println(one);
	//System.out.println(x);
	//
	//String s = one.toString();
	//BidSet bs = readBidSetFromString(s);
	//System.out.println(bs);
	//
	//Vector v = new Vector();
	//v.addElement(one);
	//v.addElement(one);
	//System.out.println(v); 
	//
	//s = v.toString();
	
//       Vector newBidCombos = BidSet.chooseOneCombos("spam", v);
//       System.out.println("spam"+ " new bid combos: " + newBidCombos);
//
//       BidSet bs = BidSet.makeSingleBidsFromBidSets("spam", newBidCombos);
//       System.out.println("spam" + " bs: " + bs);
////       ConstraintSet cs = (ConstraintSet)ht.get( (String) Util.getValue(Plist, "constraintName") );       
////       System.out.println("spam" + " constraints: " + cs);
////
////       bs.filterUsingConstraints(cs);
       BidUtility bu = BidUtility.readBidUtilityFromString("{cost_exp=-1, DSr_exp=2, DSv_exp=0, DSg_exp=0, length_exp=0, RV=1}");
//       System.out.println(" bu: " + bu);
//       bs = bs.getNHighestUtility(3, bu);
//       System.out.println("spam" + " bs: " + bs);
////       bs.setOwner("spam");
////       Local.setBidsMade(bs);//stores bid=bidset source pairs in Agent
////       System.out.println("spam" + " new bid set: " + bs.toString());

//manageFailurevAll: 
	//Vector vAll = readBidSetVector("[[{cost_exp=4, RV=[[] 0], DSv_exp=[[0 0 0][0 0 0]], length_exp=10, symbol=IBM, DSr_exp=[[4 4 1][0 0 0]], DSg_exp=[[0 0 0][0 0 0]], U=null, bidder=CR}, {topic=r, length_exp=12, cost_exp=2, DSr_exp=[[2 2 1][0 0 0]], DSg_exp=[[0 0 0][0 0 0]], DSv_exp=[[0 0 0][0 0 0]], RV=[[] 0], symbol=IBM, U=30.64850721970827, bidder=CR}, owner=CR], [{topic=r, length_exp=10, cost_exp=0, DSr_exp=[[1 1 1][0 0 0]], time_exp=1, DSg_exp=[[0 0 0][0 0 0]], DSv_exp=[[0 0 0][0 0 0]],RV=[[] 0], symbol=IBM, U=30.619822128032375, bidder=DR}, owner=DR]]");

	//System.out.println("vAll old: " + vAll);
	    //manageFailure oldBids: 
	//BidSet sourceBids = readBidSetFromString("[{topic=r, length_exp=10, cost_exp=0, DSr_exp=[[1 1 1][0 0 0]], time_exp=1, DSg_exp=[[0 0 0][0 0 0]], DSv_exp=[[0 0 0][0 0 0]], RV=[[] 0], symbol=IBM, U=30.619822128032375, bidder=DR}, {topic=r, length_exp=12, cost_exp=2, DSr_exp=[[2 2 1][0 0 0]], DSg_exp=[[0 0 0][0 0 0]], DSv_exp=[[0 0 0][0 0 0]], RV=[[] 0], symbol=IBM, U=30.64850721970827, bidder=CR}, owner=DR]");
	BidSet sourceBids = readBidsFromFile("bidfile");
	System.out.println(bu.toString());

	sourceBids.setEachBidUtility(bu);

	System.out.println("sourceBids: " + sourceBids);

	System.out.println("sourceBids: " + sourceBids.clone());

	Vector v = sourceBids.getNSortedHighestUtility(2, bu);
	System.out.println("0 is: " + (Bid)v.get(0));
	    
	    
    }
} // BidSet



//
//    //failedBids: 
//    Bid b = Bid.readBidFromString("{topic=r, length_exp=12, cost_exp=2, DSr_exp=[[2 2 1][0 0 0]], DSg_exp=[[0 0 0][0 0 0]], DSv_exp=[[0 0 0][0 0 0]], RV=[[] 0], symbol=IBM, U=30.64850721970827, bidder=CR}");
//    Vector failedBids = new Vector();
//    failedBids.add(b);
//    Vector newFailed = new Vector();
//
//    System.out.println("failedBids: " + failedBids);
//	    
//	    //All BidSets from bidding process are in vAll
//	    Iterator oldBidSets = vAll.iterator();
//	    BidSet currOldBidSet;
//	    Bid old, failed, sourceBid;
//	    Iterator failedBI;
//	    Iterator sources;
//	    while(oldBidSets.hasNext()){
//		currOldBidSet = (BidSet)oldBidSets.next();
//		Iterator currOldBidSetIter = currOldBidSet.iterator();
//		while(currOldBidSetIter.hasNext()){
//		    old = (Bid)currOldBidSetIter.next();
//		    failedBI = failedBids.iterator();
//		    while(failedBI.hasNext()){
//			failed = (Bid)failedBI.next();
//			if (failed.equals(old))
//			    newFailed.add(old);
//		    }
//		}
//		System.out.println("newFailed " + newFailed);
//		for(int i=0; i<newFailed.size(); i++)
//		    System.out.println("rem: " + currOldBidSet.remove(newFailed.elementAt(i)));
//		newFailed.clear();
//		currOldBidSetIter = currOldBidSet.iterator();
//		while(currOldBidSetIter.hasNext()){
//		    old = (Bid)currOldBidSetIter.next();
//		    sources = sourceBids.iterator();
//		    while(sources.hasNext()){
//			sourceBid = (Bid)sources.next();
//			System.out.println("comparing: " + sourceBid + " to old: " + old);
//			
//			if (old.equals(sourceBid)){
//			    System.out.println("found " + sourceBid + " in " + old);
//			    old.put("cost_exp", "0");
//			    old.put("time_exp", "0");
//			}
//		    }
//		}
//	    }
//	    System.out.println("vAll new: " + vAll);
