import java.util.*;
/**
 * BidUtility.java
 *
 * BidUtility is the set of coefficients that will be applied to a bid
 * to determine the utility of the bid. Bid represents expected values
 * of attributes (before execution) or real values of attributes
 * (after execution).
 *
 * The actual utility computation using this class takes place in
 * Bid.java.
 *
 * All term coeffs are scaled in BidUtility.java before they are applied.
 *
 * Created: Tue Jul 29 22:21:40 2003
 *
 * @author <a href="mailto:terry@Terry-Harveys-Computer.local.">Terry Harvey</a>
 * @version 1.0
 */
public class BidUtility extends Hashtable{

    static boolean DEBUG = false;
    static int nameSuffix = 0;

    public BidUtility() {

    } // BidUtility constructor

    public void setAttributeCoeff(String att, double d){
	put(att, Double.toString(d));
    }

    public void setAttributeCoeff(String att, String d){
	put(att, d);
    }

    static public BidUtility readBidUtilityFromString(String s){
	String key;
	String value;
	BidUtility h = new BidUtility();
	StringTokenizer st = new StringTokenizer(s, "{}, =");
	while (st.hasMoreTokens()){
	    key = st.nextToken();
	    value = st.nextToken();
	    h.put(key,value);
	}
	if ( ! h.containsKey("name") )
	    h.put("name", getUniqueName());
	return h;
    }

    static public BidUtility readBidUtility(StringTokenizer st){
	String key, eq;
	String value;
	BidUtility h = new BidUtility();
	//StringTokenizer st = new StringTokenizer(s, "{}, =");
	//Pattern p = Pattern.compile("[, {=]");
	while (st.hasMoreTokens()){
	    key = st.nextToken(", {}=");
	    if (DEBUG)System.out.println(key + " ");
	    
	    if (key.equals("}")) break;
	    else if ( key.matches("[, {]") )
		continue;
	    eq = st.nextToken();
	    value = st.nextToken();
	    h.put(key,value);
	}
	if ( ! h.containsKey("name") )
	    h.put("name", getUniqueName());
	return h;
    }

    static String getUniqueName(){
	nameSuffix += 1;
	return "bu" + nameSuffix;
    }

    /**
     * Return a clone of this, but with all coeffs except DS and RV
     * set to zero. For comparing texts after length, cost etc have
     * already been determined (during bid process). If more terms are
     * added to Utility, this should be rewritten to only zero length,
     * cost, time, etc.
     **/
    public BidUtility dsOnlyClone(){//RENAME THIS FCN for zeroing rv too
	BidUtility result = new BidUtility();
	Enumeration e = keys();
	String next = null;
	double temp = 1;
	while (e.hasMoreElements()){
	    next = (String)e.nextElement();
	    if (next.endsWith("coeff")){
		if ( ! (next.startsWith("DS") || next.startsWith("RV")))
		    result.put(next, "0");
		else
		    result.put(next, this.get(next));
	    }
	    else
		result.put(next, this.get(next));
	}
	result.put("name", getUniqueName());
	return result;
    }//dsOnlyClone()


    /**
     * Return a clone of this, but with all coeffs except for listed
     * prefixes set to zero. For comparing bids only by looking at the
     * utility of a subset of attributes.
     **/
    public BidUtility subsetAttrClone(Vector attributePrefixes){
	BidUtility result = new BidUtility();
	Enumeration eK = keys();
	String next = null, nextAtt = null;
	double temp = 1;
	while (eK.hasMoreElements()){
	    next = (String)eK.nextElement();
	    if (next.endsWith("coeff")){

		boolean contains = false;
		Enumeration eAttPre = attributePrefixes.elements();
		while (eAttPre.hasMoreElements()){
		    nextAtt = (String)eAttPre.nextElement();
		    if (next.startsWith(nextAtt))
			contains = true;
		}
		if ( ! contains ){ 
		    result.put(next, "0");
		}
	    }
	}
	result.put("name", getUniqueName());
	return result;
    }//subsetAttrClone()

    /**
     * Return a clone of this, but with DS coeffs reversed; assumes
     * coeffs are 1,5, and some middle that doesn't get changed. Used
     * for doing large test runs with varying U fcn.
     **/
    public BidUtility reverseDSUclone(){
	BidUtility result = new BidUtility();
	Enumeration e = keys();
	String next = null;
	double temp = 1;
	while (e.hasMoreElements()){
	    next = (String)e.nextElement();
	    if (next.startsWith("DS") &&
		next.endsWith("coeff")){
		temp = Double.parseDouble((String)get(next));
		switch ((int)temp) {
		case 1: temp = 5; break;
		case 5: temp = 1; break;
		}
		result.put(next, Double.toString(temp));
	    }
	}
	result.put("name", getUniqueName());
	return result;
    }//reverseDSUclone()

    public String hiCoeff(){
	Enumeration e = keys();
	String next, bestName = null;
	double best = 0, temp;
	while (e.hasMoreElements()){
	    next = (String)e.nextElement();
	    if (next.startsWith("DS") &&
		    next.endsWith("coeff")){
		temp = Double.parseDouble((String)get(next));
		if (temp > best){
		    best = temp;
		    bestName = next;
		}
	    }
	}
	if (bestName != null) return bestName;
	else return null;
    }

    public double sumCoeffs(){
	Enumeration e = keys();
	String next;
	double sum = 0, temp;
	while (e.hasMoreElements()){
	    next = (String)e.nextElement();
	    if ( next.endsWith("coeff"))
		sum += Double.parseDouble((String)get(next));
	}
	return sum;
    }

    /**
     * Returns a random alternative to this utility function. Only
     * varies DS coeffs.
     **/
     public BidUtility randomized(int trials, int scale){
	//be careful - order not consistent

	 double distance, bestDistance = 0;

	Set s = this.keySet();
	int length = s.size();
	///names in array will be in constant order
	String[] names = new String[length];
	s.toArray(names);
	
	Object[] original = new Object[length];
	double[] originalNums = new double[length];
	Object[] best = original;
	double[] bestNums = originalNums;
	Object[] next = new Object[length];
	double[] nextNums = new double[length];

	//init array of original values from this BidU, for comparison
	for(int i = 0; i<length; i++){
	    if ( ((String)names[i]).startsWith("DS") &&
		  ((String)names[i]).endsWith("coeff") ){
		original[i] = this.get(names[i]);
		originalNums[i] = Double.parseDouble((String)original[i]);
	    }
	    else{
		original[i] = this.get(names[i]);
		originalNums[i] = 0;
	    }
	}

	if (DEBUG) {
	    for(int i = 0; i < length; i++)
		System.out.println(names[i] + "\t" + original[i] + "\t" + originalNums[i]);
	    System.out.print("\n");
	}

	for (int j = 0; j < trials; j++){
	
	    for(int i = 0; i<length; i++){//REPLACE! store indices that need to change
		if ( ((String)names[i]).startsWith("DS") &&
		      ((String)names[i]).endsWith("coeff") ){
		    nextNums[i] = Math.random() * scale;
		    next[i] = Double.toString(nextNums[i]);
		}
		else{
		    next[i] = this.get(names[i]);
		    nextNums[i] = 0;
		}
	    }

	    if (DEBUG) {
		for(int i = 0; i < length; i++)
		    System.out.println(names[i] + "\t" + next[i] + "\t" + nextNums[i]);
		System.out.print("\n");
	    }
	
	    distance = 1 - CosineDistance.twoArraySimilarity(originalNums, nextNums);
	    if (distance > bestDistance){
		bestDistance = distance;
		best = next;
		bestNums = nextNums;
	    }
	}

	BidUtility bu = new BidUtility();
	for (int i = 0; i < length; i++){
	    bu.put(names[i], best[i]);
	}
	if (DEBUG) System.out.println(bu);
	bu.put("name", getUniqueName());
	return bu;
    }

    public static void main(String[] args){

	
       BidUtility bu, bu2;
       String sbu = "{length_coeff=10, DSg_coeff=1, DSv_coeff=5, DSv_fcn=linear, length_target=85, DSg_target=5, RV_coeff=1, DSr_target=5, DSg_fcn=linear, DSr_fcn=linear, cost_coeff=10, cost_fcn=zeroNormal, DSv_target=5, cost_target=15, length_fcn=zeroNormal, DSr_coeff=3}";

       bu = BidUtility.readBidUtility(new StringTokenizer(sbu,", {}=" ,true));
       System.out.println(bu);
       //String temp = bu.toString();
       //bu = BidUtility.readBidUtilityFromString(temp);
       //System.out.println(bu);
       //
       //BidUtility DSbu = bu.dsOnlyClone();
       //System.out.println(DSbu);

       //bu = bu.randomized(5, 10);

       //Bid valueBid = Bid.readBidFromString("{topic=v, length_exp=34.0, DStot_exp=[[4.05 5.05 2][8 8 1]], cost_exp=8.0, DSr_exp=[[0 0 0][0 0 0]], DSg_exp=[[0 0 0][0 0 0]], DSv_exp=[[4.05 5.05 2][8 8 1]], RV=[[2 5] 3], symbol=DCX, U=31.255448195700467, bidder=Value}");
       //System.out.println(valueBid);
       //valueBid.getUtility(DSbu);   
       //System.out.println(valueBid);    
       //
       //Bid riskBid = Bid.readBidFromString("{topic=r, length_exp=40.0, DStot_exp=[[2 3 2][10 11 2]], cost_exp=6.0, time_exp=2.0, DSr_exp=[[2 3 2][10 11 2]], DSg_exp=[[0 0 0][0 0 0]], DSv_exp=[[0 0 0][0 0 0]], symbol=DCX, RV=[[5 2 5] 4], U=26.982997659911568, bidder=Risk}");       
       //
       //System.out.println(riskBid);
       //riskBid.getUtility(DSbu);   
       //System.out.println(riskBid);   

       Vector atts = new Vector();
       atts.add("len");
       atts.add("cos");
       System.out.println("sans atts: " + atts + "\n" + bu);
       bu = bu.subsetAttrClone(atts);
       System.out.println("sans atts: " + atts + "\n" + bu);
       System.out.println("sum: " + bu.sumCoeffs());
    }

} // BidUtility
