// Evan Steinberger
// CISC 466
// Spring 2007

import java.util.ArrayList;
import java.util.Calendar;

/**
 * The BlindAnalyzer is the central analysis tool for determining and evaluating values.  
 * After an array of Grabbers performs their data mining and parsing, the BlindAnalyzer is constructed.  
 * The BlindAnalyzer knows nothing about the reliability of each Grabber or any historical data whatsoever 
 * (hence the term, BlindAnalyzer).
 * 
 * @author Evan Steinberger
 */
public class BlindAnalyzer
{
	
	// In SINGLEPASS Sweep Mode, more sources will be used, so the ZSCORE_SENS variable should be adjusted
	// accordingly.  In the NPASS Sweep Mode, less sources will be used but the ones that are will
	// be more consistent.
	// Also, the percentages should be adjusted in order to put more emphasis on the number of soures used
	// rather than the STD.
	private static enum SWEEP_MODE { SINGLEPASS, NPASS };

	/* THESE VARIABLES COMPRISE A "PROFILE".  THESE DEFAULT NUMBERS REPRESENT THE
		"STANDARD" PROFILE - which will be used if no valid profile is specified
	------------------------------------------------------------------------------ */
	private String profileName;
	private SWEEP_MODE CurrentMode = SWEEP_MODE.NPASS;
	// Increase this variable to allow a greater number of possible outliers included
	// Lowering this number NPASS mode will result in significantly greater
	// 2nd to n order outliers and reduce confidence
	private double ZSCORE_SENS = 1.21;

	// Rounding also affects the final numbers
	private int DECIMAL_PLACE = 3;

	// Confidence Ratios - must total to 1.0
	// Increasing the value of sources used will put more emphasis on building large lists of data
	// Increasing the value of STD will put more emphasis on consistency of data
	private double SOURCES_USED_PERCENTAGE = 0.4;
	private double STD_PERCENTAGE = 0.6;

	/* ------------------------------------------------------------------------------ */

	private String stockSymbol = "";
	
	// ArrayLists for StockInfoGrabbers
	private ArrayList grabbers;
	public ArrayList cr_outliers, pe_outliers, de_outliers, re_outliers;

	// Means
	public double cr_mean, pe_mean, de_mean, re_mean;

	// Counts of Sources actually used
	public int cr_count, pe_count, de_count, re_count;

	// Standard Deviations
	public double cr_std, pe_std, de_std, re_std;

	// Confidence Percentages
	public double cr_confidence, pe_confidence, de_confidence, re_confidence;

	// So the grabber analayzer doesn't have to read historical data again
	public String orig;

	/**
	 * Constructs the BlindAnalyzer with a standard profile
	 */
	public BlindAnalyzer()
	{
		this.profileName = "standard";
		this.init();
	}

	/**
	 * Constructs the BlindAnalyzer with a specified profile
	 * 
	 * @param profileName the profile to use
	 */
	public BlindAnalyzer(String profileName)
	{
		// if profile can't be loaded be sure to set these vars to default
		if (!this.loadProfile(profileName))
		{
			this.profileName = "standard";
			CurrentMode = SWEEP_MODE.NPASS;
			ZSCORE_SENS = 1.21;
			DECIMAL_PLACE = 3;
			SOURCES_USED_PERCENTAGE = 0.4;
			STD_PERCENTAGE = 0.6;
		}
		this.profileName = profileName;
		this.init();
	}
	
	/**
	 * Loads a profile's settings into BlindAnalyzer
	 * 
	 * @param profileName the profile to use
	 * @return true if successful, false if not
	 */
	private boolean loadProfile(String profileName)
	{
		String content = ReadData.fromFile(profileName + ".blindprofile");
		if (content.equals("!!ERROR"))
			return false;

		try
		{
			String tMode, tStdGrace, tDec, tSUP, tSP;
			int startIndex;

			startIndex = content.indexOf("CurrentMode");
			content = content.substring(startIndex);
			tMode = content.substring(12, content.indexOf("\r\n"));

			startIndex = content.indexOf("ZSCORE_SENS");
			content = content.substring(startIndex);
			tStdGrace = content.substring(12, content.indexOf("\r\n"));

			startIndex = content.indexOf("DECIMAL_PLACE");
			content = content.substring(startIndex);
			tDec = content.substring(14, content.indexOf("\r\n"));

			startIndex = content.indexOf("SOURCES_USED_PERCENTAGE");
			content = content.substring(startIndex);
			tSUP = content.substring(24, content.indexOf("\r\n"));

			startIndex = content.indexOf("STD_PERCENTAGE");
			content = content.substring(startIndex);
			tSP = content.substring(15, content.indexOf("\r\n"));

			if (tMode.equals("NPASS"))
				CurrentMode = SWEEP_MODE.NPASS;
			else if (tMode.equals("SINGLEPASS"))
				CurrentMode = SWEEP_MODE.SINGLEPASS;
			else
				return false;

			ZSCORE_SENS = Double.parseDouble(tStdGrace);
			System.out.println(ZSCORE_SENS);
			DECIMAL_PLACE = Integer.parseInt(tDec);
			SOURCES_USED_PERCENTAGE = Double.parseDouble(tSUP);
			STD_PERCENTAGE = Double.parseDouble(tSP);
			System.out.println("Loaded Blind profile: " + profileName);
			return true;
		}
		catch (Exception ex)
		{
			System.out.println("Blind Profile " + profileName + " failed to load: " + ex.toString());
			return false;
		}
		
	}
	
	/**
	 * Initializes all values for the BlindAnalyzer
	 */
	private void init()
	{
		grabbers = new ArrayList();
		cr_outliers = new ArrayList();
		pe_outliers = new ArrayList();
		de_outliers = new ArrayList();
		re_outliers = new ArrayList();
		cr_mean = pe_mean = de_mean = re_mean = 0;
		cr_std = pe_std = de_std = re_std = 0;
		cr_count = pe_count = de_count = re_count = 0;
		cr_confidence = pe_confidence = de_confidence = re_confidence = 0;
	}
	
	/**
	 * Adds a StockInfoGrabber to the BlindAnalyzer for analysis
	 * 
	 * @param sig the StockInfoGrabber to Add
	 */
	public void add(StockInfoGrabber sig)
	{
		grabbers.add(sig);
		// TEMP... should be fixed later
		this.stockSymbol = sig.getStockSymbol();
	}
	
	/**
	 * Perform all blind analysis techniques on all StockInfoGrabbers
	 */
	public void analyze()
	{
		System.out.println("Blind Analysis started ... ");
		this.calculateMeans();
		this.calculateSTDS();
		cr_confidence = buildConfidence(cr_mean, cr_std, cr_count, "CR", cr_outliers);
		pe_confidence = buildConfidence(pe_mean, pe_std, pe_count, "PE", pe_outliers);
		de_confidence = buildConfidence(de_mean, de_std, de_count, "DE", de_outliers);
		re_confidence = buildConfidence(re_mean, re_std, re_count, "RE", re_outliers);
		this.writeHistoricalContent();
		System.out.println("Blind Analysis complete");
	}
	
	/**
	 * Write out historical data to a history file
	 * 
	 * @return true if successful, false if not
	 */
	private boolean writeHistoricalContent()
	{
		orig = ReadData.fromFile(this.getClass().getName() + ".txt");
		if (orig.equals("!!ERROR"))
			orig = "";
		else
			orig += "\r\n\r\n";

		Calendar now = Calendar.getInstance();
		orig += this.stockSymbol + " " + "date" + " " + (now.get(Calendar.MONTH) + 1) + "/" + now.get(Calendar.DAY_OF_MONTH) + "/" + now.get(Calendar.YEAR) + " " + now.get(Calendar.HOUR_OF_DAY) + ":" + now.get(Calendar.MINUTE) + "\r\n";
		orig += this.stockSymbol + " " + "total sources" + " " + grabbers.size() + "\r\n";
		orig += this.stockSymbol + " " + "profile" + " " + this.profileName + "\r\n";
		orig = this.buildSingleHistory(orig, cr_mean, cr_confidence, cr_std, cr_count, cr_outliers, "cr");
		orig = this.buildSingleHistory(orig, pe_mean, pe_confidence, pe_std, pe_count, pe_outliers, "pe");
		orig = this.buildSingleHistory(orig, de_mean, de_confidence, de_std, de_count, de_outliers, "de");
		orig = this.buildSingleHistory(orig, re_mean, re_confidence, re_std, re_count, re_outliers, "re");

		return WriteData.toFile(this.getClass().getName() + ".txt", orig);
	}

	/**
	 * Build a single history entry
	 * 
	 * @param orig the original history text
	 * @param mean the mean to write out
	 * @param confidence the confidence to write out
	 * @param std the standard deviation to write out
	 * @param count the count of sources to write out
	 * @param outliers the arraylist of outliers
	 * @param which which ratio is being examined
	 * @return the string data of the new history entry
	 */
	private String buildSingleHistory(String orig, double mean, double confidence, double std, int count, ArrayList outliers, String which)
	{
		orig += this.stockSymbol + " " + which + " mean" + " " + mean + "\r\n";
		orig += this.stockSymbol + " " + which + " confidence" + " " + confidence + "\r\n";
		orig += this.stockSymbol + " " + which + " std" + " " + std + "\r\n";
		orig += this.stockSymbol + " " + which + " sources used" + " " + count + "\r\n";
		orig += this.stockSymbol + " " + which + " outliers" + " ";
		if (outliers.size() > 0)
		{
			for (Object o : outliers)
				orig += ((StockInfoGrabber)o).getClass().getName() + ",";
			orig = orig.substring(0, orig.length()-1);
		}
		else
			orig+= "none";
		orig += "\r\n";
		return orig;
	}
	
	/**
	 * Builds a confidence level for a particular ratio.
	 * 
	 * @param mean the arithmetic mean
	 * @param std the standard deviation
	 * @param count the number of sources used
	 * @param whichData which Ratio is being examined
	 * @param outliers an arraylist of outliers found
	 * @return a confidence level
	 */
	private double buildConfidence(double mean, double std, int count, String whichData, ArrayList outliers)
	{
		while (findOutliers(mean, std, whichData, outliers))
		{
			if (whichData=="CR")
			{
				this.calculateMeans();
				this.cr_std = this.round(this.calcSTD(cr_mean, cr_count, cr_outliers, "CR"));
				mean = cr_mean;
				std = cr_std;
				count = cr_count;
			}
			else if (whichData=="PE")
			{
				this.calculateMeans();
				this.pe_std = this.round(this.calcSTD(pe_mean, pe_count, pe_outliers, "PE"));
				mean = pe_mean;
				std = pe_std;
				count = pe_count;
			}
			else if (whichData=="DE")
			{
				this.calculateMeans();
				this.de_std = this.round(this.calcSTD(de_mean, de_count, de_outliers, "DE"));
				mean = de_mean;
				std = de_std;
				count = de_count;
			}
			else if (whichData=="RE")
			{
				this.calculateMeans();
				this.re_std = this.round(this.calcSTD(re_mean, re_count, re_outliers, "RE"));
				mean = re_mean;
				std = re_std;
				count = re_count;
			}
			if (CurrentMode == SWEEP_MODE.SINGLEPASS)
				break;
		}
		
		double con=0;

		// For every source lost, confidence is reduced by some percentage
		// if there is only one source we cannot hope to verify it.
		if (count == 1 || count == 0)
			con+= 0;
		else
		{
			double usedRatio= this.round((double)((double)count / (double)grabbers.size()) * 100 * SOURCES_USED_PERCENTAGE);
			con+= usedRatio;
		}
		
		// Add in the standard deviation to the confidence by
		// 100 - (STD * 100)
		if (count != 0)
			con+= ((100 - (std * 100)) * STD_PERCENTAGE);
		return Math.abs(con);
	}
	
	/**
	 * Find outliers for a set of StockInfoGrabbers
	 * 
	 * @param mean the mean of the data set
	 * @param std the standard deviation of the data set
	 * @param whichData which ratio is being examined
	 * @param outliers the list of outliers
	 * @return true if outliers were found, false if not
	 */
	private boolean findOutliers(double mean, double std, String whichData, ArrayList outliers)
	{
		boolean b = false;
		for (Object o : grabbers)
		{
			StockInfoGrabber sig = (StockInfoGrabber)o;
			double v=-1;
			v = sig.getValue(whichData);
			if (v==-1 || outliers.contains(sig))
				continue;
			
			// New - normalized Data with z scores
			
			//double high = this.round((mean+std+.01)) ;
			//double low = this.round((mean-std-.01));

			double zScore = (v - mean) / std;
			zScore = this.round(zScore);
			
			//if (v> high || v < low)
			if (Math.abs(zScore) > ZSCORE_SENS)
			{
				b = true;
				outliers.add(sig);
			}
		}
		return b;
	}

	/**
	 * Calculates the arithmetic mean for each ratio
	 */
	private void calculateMeans()
	{
		cr_mean = 0;
		pe_mean = 0;
		de_mean = 0;
		re_mean = 0;
		cr_count = 0;
		pe_count = 0;
		de_count = 0;
		re_count = 0;
		for (Object o : grabbers)
		{
			StockInfoGrabber sig = (StockInfoGrabber)o;
			if (sig.getCurrentRatio() != -1 && !cr_outliers.contains(sig))
			{
				cr_mean+= sig.getCurrentRatio();
				cr_count++;
			}
			if (sig.getPeTTM() != -1 && !pe_outliers.contains(sig))
			{
				pe_mean+= sig.getPeTTM();
				pe_count++;
			}
			if (sig.getDeTotal() != -1 && !de_outliers.contains(sig))
			{
				de_mean+= sig.getDeTotal();
				de_count++;
			}
			if (sig.getReTTM() != -1 && !re_outliers.contains(sig))
			{
				re_mean+= sig.getReTTM();
				re_count++;
			}
		}
		
		cr_mean/= cr_count;
		pe_mean/= pe_count;
		de_mean/= de_count;
		re_mean/= re_count;
		cr_mean = this.round(cr_mean);
		pe_mean = this.round(pe_mean);
		de_mean = this.round(de_mean);
		re_mean = this.round(re_mean);
	}



	/*
	private double calcMean(String whichData)
	{
		double mean;
		int count;
		for (Object o : grabbers)
		{
			StockInfoGrabber sig = (StockInfoGrabber)o;
			if (sig.getData(whichData) != -1)
		}
	}*/
	
	/**
	 * Calculates the standard deviation for each ratio
	 */
	private void calculateSTDS()
	{
		cr_std = pe_std = de_std = re_std = 0;
		this.cr_std = this.round(this.calcSTD(cr_mean, cr_count, cr_outliers, "CR"));
		this.pe_std = this.round(this.calcSTD(pe_mean, pe_count, pe_outliers, "PE"));
		this.de_std = this.round(this.calcSTD(de_mean, de_count, de_outliers, "DE"));
		this.re_std = this.round(this.calcSTD(re_mean, re_count, re_outliers, "RE"));
	}

	/**
	 * Calculates the standard deviation for one ratio
	 * 
	 * @param mean the mean of the data set
	 * @param size the size of the data set
	 * @param outliers the outliers of the data set
	 * @param whichData which ratio is being examined
	 * @return the standard deviation of the data set
	 */
	private double calcSTD(double mean, int size, ArrayList outliers, String whichData)
	{
		double sum = 0;
		for (Object o : grabbers)
		{
			StockInfoGrabber sig = (StockInfoGrabber)o;
			double v=-1;
			v = sig.getValue(whichData);
			if (v==-1 || outliers.contains(sig))
				continue;
			v-= mean;
			sum += (v*v);
		}
		return Math.sqrt(sum / (size - 1) );
	}

	/**
	 * Rounds a value.
	 * 
	 * @param value the value to round
	 * @return the rounded value
	 */
	private double round(double value) 
	{
		int decimalPlace = DECIMAL_PLACE;
		double power_of_ten = 1;
		while (decimalPlace-- > 0)
			power_of_ten *= 10.0;
		return Math.round(value * power_of_ten) / power_of_ten;
	}
}