import java.lang.Math;
import java.lang.reflect.*;
/**
 * DistFcn.java
 *
 * Functions that determine the value of an expected value given a
 * target value. Use for calculating utility. Functions return
 * 0-1. After this is applied, a coeff from the utility fcn is also
 * applied. See BidUtility and Bid.
 *
 * Created: Tue Feb 24 10:28:17 2004
 *
 * @author <a href="mailto:terry@Terry-Harveys-Computer.local.">Terry Harvey</a>
 * @version 1.0
 */
public class DistFcn {
    
    static final boolean DEBUG = false;
    static final DistFcn example = new DistFcn(); //used to avoid creating new objects in dynamic method invocation
    static final double SIGMA_DEFAULT = 5;

    public DistFcn() {
	
    } // DistFcn constructor
    
    //sdEst is portion of mean that will be used as standard dev to affect spread of curve
    public static double fProportionalDev(double x, double mean, double sdEst){
	double sigma = sdEst * mean;  //est of standard deviation
	return  fFixedDev(x, mean, sigma);
    }

    public static double fFixedDev(double x, double mean, double sigma){

	double value, exponent, denom, maxHeight;
	
	if (sigma == 0){ 
	    sigma = SIGMA_DEFAULT;
	}
	exponent = - Math.pow( (x - mean), 2)/ (2 * Math.pow( sigma, 2));
	//if (DEBUG) System.out.println("exponent: " + exponent);
	//if (DEBUG) System.out.println("num " + (- Math.pow( (x - mean), 2)) + " denom " + (2 * Math.pow( sigma, 2)));
	
	value = Math.exp(exponent); //this is e^exponent
	//if (DEBUG) System.out.println("value: " + value);
	
	return value;
    }

    public static double normal(Double x, Double mean, Double sigma){
	return fFixedDev(x.doubleValue(), mean.doubleValue(), sigma.doubleValue());
    }
    //right half of normal curve, high at zero
    //sigma is ignored; target mean becomes sigma
    public static double zeroNormal(Double x, Double mean, Double sigma){
	return fFixedDev(x.doubleValue(), 0, mean.doubleValue());
    }
    /**
     *is left half of normal curve until mean, then is 1
     **/
    public static double normPlateauEnd(Double x, Double mean, Double sigma){
	if (x.doubleValue() >= mean.doubleValue()) return 1;
	else return fFixedDev(x.doubleValue(), mean.doubleValue(), sigma.doubleValue());
    }
    //startPlateauNormal
    /**
     * returns 1 until mean, then is right half of normal curve
     **/
    public static double startPlateauNormal(Double x, Double mean, Double sigma){
	if (x.doubleValue() <= mean.doubleValue()) return 1;
	else return fFixedDev(x.doubleValue(), mean.doubleValue(), sigma.doubleValue());
    }
    
    // y = 1 at x = target
    public static double linear(Double x, Double target, Double notUsed){
	return x.doubleValue() / target.doubleValue();
    }

    // y = 1 at x < target; y = 0 at x = 2 * target; linear between; FIX for zero target
    public static double startPlateauLinear(Double x, Double target, Double notUsed){
	if (x.doubleValue() < target.doubleValue())
	    return 1;
	else if (x.doubleValue() < (2 * target.doubleValue()))
	    return 1 - ( (x.doubleValue() - target.doubleValue()) / target.doubleValue());
	else return 0;
    }
    
    // y = 1 at x >= target
    public static double linearPlateauEnd(Double x, Double target, Double notUsed){
	if (x.doubleValue() >= target.doubleValue())
	    return 1;
	else
	    return x.doubleValue() / target.doubleValue();
    }
    // y = 0 at x >= target
    public static double linearDownSlope(Double x, Double target, Double notUsed){
	if (x.doubleValue() < target.doubleValue())
	    return 1 - (x.doubleValue() / target.doubleValue());
	else return 0;
    }

    /*
     * Call the named fcn on the args.
     */

    public static double f(String name, double x, double mean, double sigma){
    
	return f(name, new Double(x), new Double(mean), new Double(sigma));
    }
    public static double f(String name, Double x, Double mean, Double sigma){

	Method m = DistFcn.getFcn(name);
	double result;
	Object[] args = {x, mean, sigma};
	try{
	    result = ((Double)m.invoke( DistFcn.example, args)).doubleValue();//why not "this"?
	    if (DEBUG) System.out.println("f returning: " + result);
	    
	    return result;
	    
	} catch (IllegalAccessException e) {
	    System.out.println(e);
	} catch (InvocationTargetException e) {
	    System.out.println(e);
	}
	return -100;//large negative number will upset utility and be noticed?
    }

    public static Method getFcn(String name){
	Class[] params = {Double.class, Double.class, Double.class};
	Class fcns;;
	Method f;
	try {
	    fcns = Class.forName("DistFcn");
	    f = fcns.getMethod(name, params);
	    return f;
	} catch (Exception e) {
	    System.out.println("Error finding Ufcn: " + name + " " + params);
	    
	    return null;
	}
    }

    public static void main(String[] args2){

	
	System.out.println(" DistFcn.f(normal, 100,0,75));value is: " + DistFcn.f("normal", 100,0,75));
	System.out.println(" DistFcn.f(normal, 90,0,75));value is: " + DistFcn.f("normal", 90,0,75));
	System.out.println(" DistFcn.f(normal, 85,0,75));value is: " + DistFcn.f("normal", 85,0,75));
	System.out.println(" DistFcn.f(normal, 80,0,75));value is: " + DistFcn.f("normal", 80,0,75));

	System.out.println(" DistFcn.f(normal, 75,0,75));value is: " + DistFcn.f("normal", 75,0,75));
	System.out.println(" DistFcn.f(normal, 50,0,75));value is: " + DistFcn.f("normal", 50,0,75));
	System.out.println(" DistFcn.f(normal, 20,0,75));value is: " + DistFcn.f("normal", 20,0,75));
	System.out.println(" DistFcn.f(normal, 20,0,75));value is: " + DistFcn.f("zeroNormal", 20,75,150));

    }
} // DistFcn
