// Evan Steinberger
// CISC 466
// Spring 2007

import java.util.ArrayList;
import java.util.Calendar;

/**
 * The GrabberAnalyzer evaluates the overall worth of each Grabber.  
 * It determines which Grabbers are the most and least reliable for the four key values of CR, PE, RE, and DE.  
 * It examines both the BlindAnalyzers history file as well as each of the Grabbers own history files to perform its calculations and analysis.  
 * It also provides the average confidence levels for all browsers currently being used.  
 * After performing calculations, the GrabberAnalyzer stores its results into a total history file associated with it.  
 * It also stores a history file of analytical data for each individual Grabber.  
 * The GrabberAnalyzer would also be the basis for rule changes and Grabber restructuring, because it keeps track of how often each Grabber fails.
 * 
 * @author Evan Steinberger
 */

public class GrabberAnalyzer
{
	private ArrayList blindEntries;
	private String rawBlindText;

	private double uniqueBlindRatio;
	private ArrayList uniqueBlinds;

	private ArrayList grabbersInfo;
	
	// These numbers comprise a profile
	private double ERROR_PERCENTAGE = .4;
	private double OUTLIER_PERCENTAGE = .4;
	private double EMPHASIS_UNIQUE_SYMBOL_BLIND = .05;
	private double EMPHASIS_UNIQUE_SYMBOL_HISTORY = .15;

	public String crBest, peBest, deBest, reBest;
	public String crWorst, peWorst, deWorst, reWorst;

	public double crAvg, peAvg, deAvg, reAvg;

	String profileName = "standard";

	/**
	 * Constructs a GrabberAnalyzer
	 * 
	 * @param grabbers an array of StockInfoGrabbers
	 * @param blindText raw BlindAnalyzer history text
	 * @param profileName the profile to use
	 */
	public GrabberAnalyzer(ArrayList grabbers, String blindText, String profileName)
	{
		// if profile can't be loaded be sure to set these vars to default
		if (!this.loadProfile(profileName))
		{
			this.profileName = "standard";
			ERROR_PERCENTAGE = .4;
			OUTLIER_PERCENTAGE = .4;
			EMPHASIS_UNIQUE_SYMBOL_BLIND = .05;
			EMPHASIS_UNIQUE_SYMBOL_HISTORY = .15;
		}
		this.profileName = profileName;
		
		grabbersInfo = new ArrayList();
		for (Object o : grabbers)
		{
			StockInfoGrabber sig = (StockInfoGrabber)o;
			grabbersInfo.add(new GrabberEntry(sig.getClass().getName()));
		}
		
		// Get outlier information from blind outliers
		System.out.println("Adding BlindEntries to GrabberAnalyzer ... ");
		blindEntries = new ArrayList();
		uniqueBlinds = new ArrayList();

		this.rawBlindText = blindText;
		if (this.rawBlindText.equals("!!ERROR"))	
			this.rawBlindText = "";
		else
		{
			String parts[] = this.rawBlindText.split("\r\n\r\n\r\n");
			for (String s : parts)
			{
				BlindEntry be = new BlindEntry(s);
				if (!uniqueBlinds.contains(be.stockSymbol))
					uniqueBlinds.add(be.stockSymbol);
				this.blindEntries.add(be);
			}
		}
		System.out.println("Finished Adding BlindEntries to GrabberAnalyzer");

		System.out.println("Adding GrabberHistoryEntries to GrabberAnalyzer ... ");
		for (Object o : grabbersInfo)
		{
			GrabberEntry ge = (GrabberEntry)o;
			String rawHistoryText = ReadData.fromFile( ge.grabberName + ".txt" );
			if (rawHistoryText.equals("!!ERROR"))
				continue;
			String parts[] = rawHistoryText.split("\r\n\r\n\r\n");
			for (String s : parts)
			{
				GrabberHistoryEntry ghe = new GrabberHistoryEntry(s);
				if (!ge.uniqueGrabbers.contains(ghe.stockSymbol))
					ge.uniqueGrabbers.add(ghe.stockSymbol);
				ge.historyEntries.add(ghe);
			}
		}
		System.out.println("Finished Adding GrabberHistoryEntries to GrabberAnalyzer");
	}
	
	/**
	 * Performs all analysis.
	 */
	public void analyze()
	{
		this.countOutliers();
		this.countErrors();
		this.calculateConfidence();
		this.calculateBest();
		this.calculateWorst();
		this.writeHistoricalContent();
	}

	/**
	 * Calculates which StockInfoGrabber is the best choice for each ratio.
	 */
	private void calculateBest()
	{
		crBest = "none";
		peBest = "none";
		deBest = "none";
		reBest = "none";
		double tempCR, tempPE, tempDE, tempRE;
		tempCR = tempPE = tempDE = tempRE = 0;
		for (Object o : grabbersInfo)
		{
			GrabberEntry ge = (GrabberEntry)o;
			if (ge.cr_con > tempCR)
			{
				tempCR = ge.cr_con;
				crBest = ge.grabberName;
			}

			if (ge.pe_con > tempPE)
			{
				tempPE = ge.pe_con;
				peBest = ge.grabberName;
			}

			if (ge.de_con > tempDE)
			{
				tempDE = ge.de_con;
				deBest = ge.grabberName;
			}

			if (ge.re_con > tempRE)
			{
				tempRE = ge.re_con;
				reBest = ge.grabberName;
			}
		}
	}
	
	/**
	 * Calculates which StockInfoGrabber is the worst choice for each ratio.
	 */
	private void calculateWorst()
	{
		crWorst = "none";
		peWorst = "none";
		deWorst = "none";
		reWorst = "none";
		double tempCR, tempPE, tempDE, tempRE;
		tempCR = tempPE = tempDE = tempRE = 100;
		for (Object o : grabbersInfo)
		{
			GrabberEntry ge = (GrabberEntry)o;
			if (ge.cr_con < tempCR)
			{
				tempCR = ge.cr_con;
				crWorst = ge.grabberName;
			}

			if (ge.pe_con < tempPE)
			{
				tempPE = ge.pe_con;
				peWorst = ge.grabberName;
			}

			if (ge.de_con < tempDE)
			{
				tempDE = ge.de_con;
				deWorst = ge.grabberName;
			}

			if (ge.re_con < tempRE)
			{
				tempRE = ge.re_con;
				reWorst = ge.grabberName;
			}
		}
	}
	
	/**
	 * Counts the outliers for each ratio
	 */
	private void countOutliers()
	{	
		System.out.print("Counting Grabber outliers ... ");
		for (Object o : grabbersInfo)
		{
			GrabberEntry ge = (GrabberEntry)o;
			
			for (Object p : blindEntries)
			{
				BlindEntry be = (BlindEntry)p;
				if (!be.isValid)
					continue;
				if ( be.cr_outliers.indexOf(ge.grabberName) >= 0)
					ge.cr_outlier_total++;
				if ( be.pe_outliers.indexOf(ge.grabberName) >= 0)
					ge.pe_outlier_total++;
				if ( be.de_outliers.indexOf(ge.grabberName) >= 0)
					ge.de_outlier_total++;
				if ( be.re_outliers.indexOf(ge.grabberName) >= 0)
					ge.re_outlier_total++;
			}
		}
		System.out.println("done");
	}
	
	/**
	 * Counts the errors for each ratio
	 */
	private void countErrors()
	{
		System.out.print("Counting Grabber Errors ... ");
		for (Object o : grabbersInfo)
		{
			GrabberEntry ge = (GrabberEntry)o;
			for (Object p : ge.historyEntries)
			{
				GrabberHistoryEntry ghe = (GrabberHistoryEntry)p;
				if (!ghe.isValid)
					continue;
				
				if (ghe.error.equals("Error: Invalid Symbol Provided to Data Miner."))
					continue;
				if (ghe.cr==-1)
					ge.cr_errors++;
				if (ghe.de==-1)
					ge.de_errors++;
				if (ghe.pe==-1)
					ge.pe_errors++;
				if (ghe.re==-1)
					ge.re_errors++;
				if (!ghe.error.equals("none"))
					ge.totalErrors++;
			}
		}
		System.out.println("done");
	}
	
	/**
	 * Calculates the confidence in each Grabber
	 */
	private void calculateConfidence()
	{
		System.out.print("Calculating Grabber Analysis Confidence ... ");
		this.uniqueBlindRatio = (double)this.uniqueBlinds.size() / (double)this.blindEntries.size();
		
		crAvg = peAvg = deAvg = reAvg = 0;
		for (Object o : grabbersInfo)
		{
			GrabberEntry ge = (GrabberEntry)o;
			double cr_rat = ((double)this.blindEntries.size() - (double)ge.cr_outlier_total) / (double)this.blindEntries.size();
			double pe_rat = ((double)this.blindEntries.size() - (double)ge.pe_outlier_total) / (double)this.blindEntries.size();
			double de_rat = ((double)this.blindEntries.size() - (double)ge.de_outlier_total) / (double)this.blindEntries.size();
			double re_rat = ((double)this.blindEntries.size() - (double)ge.re_outlier_total) / (double)this.blindEntries.size();

			double cr_eRat = ((double)ge.historyEntries.size() - (double)ge.cr_errors) / (double)ge.historyEntries.size();
			double pe_eRat = ((double)ge.historyEntries.size() - (double)ge.pe_errors) / (double)ge.historyEntries.size();
			double de_eRat = ((double)ge.historyEntries.size() - (double)ge.de_errors) / (double)ge.historyEntries.size();
			double re_eRat = ((double)ge.historyEntries.size() - (double)ge.re_errors) / (double)ge.historyEntries.size();

			ge.uniqueGrabberRatio = (double)ge.uniqueGrabbers.size() / (double)ge.historyEntries.size();
			double tempFullGrabberRatio = 1 - EMPHASIS_UNIQUE_SYMBOL_HISTORY;
			double tempFullBlindRatio = 1 - EMPHASIS_UNIQUE_SYMBOL_BLIND;
			
			ge.cr_con = 100 * ((ERROR_PERCENTAGE * cr_eRat) + (OUTLIER_PERCENTAGE * cr_rat) + (ge.uniqueGrabberRatio*EMPHASIS_UNIQUE_SYMBOL_HISTORY) + (this.uniqueBlindRatio*EMPHASIS_UNIQUE_SYMBOL_BLIND));
			ge.pe_con = 100 * ((ERROR_PERCENTAGE * pe_eRat) + (OUTLIER_PERCENTAGE * pe_rat) + (ge.uniqueGrabberRatio*EMPHASIS_UNIQUE_SYMBOL_HISTORY) + (this.uniqueBlindRatio*EMPHASIS_UNIQUE_SYMBOL_BLIND));
			ge.de_con = 100 * ((ERROR_PERCENTAGE * de_eRat) + (OUTLIER_PERCENTAGE * de_rat) + (ge.uniqueGrabberRatio*EMPHASIS_UNIQUE_SYMBOL_HISTORY) + (this.uniqueBlindRatio*EMPHASIS_UNIQUE_SYMBOL_BLIND));
			ge.re_con = 100 * ((ERROR_PERCENTAGE * re_eRat) + (OUTLIER_PERCENTAGE * re_rat) + (ge.uniqueGrabberRatio*EMPHASIS_UNIQUE_SYMBOL_HISTORY) + (this.uniqueBlindRatio*EMPHASIS_UNIQUE_SYMBOL_BLIND));

			crAvg += ge.cr_con;
			peAvg += ge.pe_con;
			deAvg += ge.de_con;
			reAvg += ge.re_con;
		}
		crAvg /= (double)grabbersInfo.size();
		peAvg /= (double)grabbersInfo.size();
		deAvg /= (double)grabbersInfo.size();
		reAvg /= (double)grabbersInfo.size();
		System.out.println("done");
	}
	
	/**
	 * Writes data out to a history file
	 * 
	 * @return true if successful, false if not
	 */
	private boolean writeHistoricalContent()
	{
		boolean b = true;
		Calendar now = Calendar.getInstance();
		for (Object o : grabbersInfo)
		{
			GrabberEntry ge = (GrabberEntry)o;

			// Commented out because individual grabber history files getting too big
			//String orig = ReadData.fromFile(ge.grabberName + "Analysis" + ".txt");
			//if (orig.equals("!!ERROR"))
			//	orig = "";
			//else
			//	orig += "\r\n\r\n";
			String orig="";

			orig += ge.grabberName + " date" + " " + (now.get(Calendar.MONTH) + 1) + "/" + now.get(Calendar.DAY_OF_MONTH) + "/" + now.get(Calendar.YEAR) + " " + now.get(Calendar.HOUR_OF_DAY) + ":" + now.get(Calendar.MINUTE) + "\r\n";
			orig += ge.grabberName + " profile " + this.profileName + "\r\n";
			orig += ge.grabberName + " cr outlier total " + ge.cr_outlier_total + "\r\n";
			orig += ge.grabberName + " cr error total " + ge.cr_errors + "\r\n";
			orig += ge.grabberName + " cr confidence " + ge.cr_con + "\r\n";

			orig += ge.grabberName + " pe outlier total " + ge.pe_outlier_total + "\r\n";
			orig += ge.grabberName + " pe error total " + ge.pe_errors + "\r\n";
			orig += ge.grabberName + " pe confidence " + ge.pe_con + "\r\n";

			orig += ge.grabberName + " de outlier total " + ge.de_outlier_total + "\r\n";
			orig += ge.grabberName + " de error total " + ge.de_errors + "\r\n";
			orig += ge.grabberName + " de confidence " + ge.de_con + "\r\n";

			orig += ge.grabberName + " re outlier total " + ge.re_outlier_total + "\r\n";
			orig += ge.grabberName + " re error total " + ge.re_errors + "\r\n";
			orig += ge.grabberName + " re confidence " + ge.re_con + "\r\n";

			//orig += ge.grabberName + " total confidence " + ge.grabberConfidence + "\r\n";
			orig += ge.grabberName + " total blind count " + this.blindEntries.size() + "\r\n";
			orig += ge.grabberName + " total errrors " + ge.totalErrors + "\r\n";

			b = b && WriteData.toFile( ge.grabberName + "Analysis" + ".txt", orig);
		}

		String origTotal = ReadData.fromFile("TotalGrabberAnalysis.txt");
		if (origTotal.equals("!!ERROR"))
			origTotal = "";
		else
			origTotal += "\r\n\r\n";

		origTotal += "date" + " " + (now.get(Calendar.MONTH) + 1) + "/" + now.get(Calendar.DAY_OF_MONTH) + "/" + now.get(Calendar.YEAR) + " " + now.get(Calendar.HOUR_OF_DAY) + ":" + now.get(Calendar.MINUTE) + "\r\n";
		origTotal += "profile " + this.profileName + "\r\n";

		origTotal += "cr best " + this.crBest + "\r\n";
		origTotal += "cr worst " + this.crWorst + "\r\n";
		origTotal += "cr Avg Confidence " + this.crAvg + "\r\n";

		origTotal += "pe best " + this.peBest + "\r\n";
		origTotal += "pe worst " + this.peWorst + "\r\n";
		origTotal += "pe Avg Confidence " + this.peAvg + "\r\n";

		origTotal += "de best " + this.deBest + "\r\n";
		origTotal += "de worst " + this.deWorst + "\r\n";
		origTotal += "de Avg Confidence " + this.deAvg + "\r\n";

		origTotal += "re best " + this.reBest + "\r\n";
		origTotal += "re worst " + this.reWorst + "\r\n";
		origTotal += "re Avg Confidence " + this.reAvg + "\r\n";

		b = b && WriteData.toFile("TotalGrabberAnalysis.txt", origTotal);

		return b;
	}
	
	/**
	 * Loads a GrabberProfile's settings in to GrabberAnalyzer
	 * 
	 * @param profileName the profile to load
	 * @return true if successful, false if not
	 */
	private boolean loadProfile(String profileName)
	{
		String content = ReadData.fromFile(profileName + ".grabberprofile");
		if (content.equals("!!ERROR"))
			return false;

		try
		{
			String errorPer, outlierPer, blindUnique, historyUnique;
			int startIndex;

			startIndex = content.indexOf("ERROR_PERCENTAGE");
			content = content.substring(startIndex);
			errorPer = content.substring(17, content.indexOf("\r\n"));

			startIndex = content.indexOf("OUTLIER_PERCENTAGE");
			content = content.substring(startIndex);
			outlierPer = content.substring(19, content.indexOf("\r\n"));

			startIndex = content.indexOf("EMPHASIS_UNIQUE_SYMBOL_BLIND");
			content = content.substring(startIndex);
			blindUnique = content.substring(29, content.indexOf("\r\n"));

			startIndex = content.indexOf("EMPHASIS_UNIQUE_SYMBOL_HISTORY");
			content = content.substring(startIndex);
			historyUnique = content.substring(31, content.indexOf("\r\n"));

			ERROR_PERCENTAGE = Double.parseDouble(errorPer);
			OUTLIER_PERCENTAGE = Double.parseDouble(outlierPer);
			EMPHASIS_UNIQUE_SYMBOL_BLIND = Double.parseDouble(blindUnique);
			EMPHASIS_UNIQUE_SYMBOL_HISTORY = Double.parseDouble(historyUnique);
			System.out.println("Loaded Grabber profile: " + profileName);
			return true;
		}
		catch (Exception ex)
		{
			System.out.println("Grabber Profile " + profileName + " failed to load: " + ex.toString());
			return false;
		}

	}
	
	/**
	 * GrabberEntry represents one entry of a Grabber history file
	 * 
	 * @author Evan Steinberger
	 */
	private class GrabberEntry
	{
		public String grabberName;
		public double grabberConfidence;
		public int totalErrors;
		public int cr_outlier_total, pe_outlier_total, de_outlier_total, re_outlier_total;
		public int cr_errors, pe_errors, de_errors, re_errors;
		public double cr_con, pe_con, de_con, re_con;
		private ArrayList historyEntries;
		public double uniqueGrabberRatio;
		private ArrayList uniqueGrabbers;

		/**
		 * Constructs a GrabberEntry
		 * 
		 * @param name the name of the StockInfoGrabber
		 */
		public GrabberEntry(String name)
		{
			this.grabberName=name;
			this.historyEntries = new ArrayList();
			cr_outlier_total = pe_outlier_total = de_outlier_total =re_outlier_total=0;
			cr_errors = pe_errors = de_errors = re_errors = totalErrors = 0;
			cr_con = pe_con = de_con = re_con = 0;
			grabberConfidence = 0;
			uniqueGrabberRatio =1;
			uniqueGrabbers = new ArrayList();
		}
	}
	
	/**
	 * GrabberHistoryEntry represents one entry of a GrabberAnalyzer history file
	 * 
	 * @author Evan Steinberger
	 */
	private class GrabberHistoryEntry
	{
		public String stockSymbol;
		public String date;
		public String industry;
		public double cr, de, pe, re;
		public String error;
		public boolean isValid;

		private String rawText;

		/**
		 * Constructs a GrabberHistoryEntry
		 * 
		 * @param rawT the raw text of the GrabberHistoryEntry
		 */
		public GrabberHistoryEntry(String rawT)
		{
			this.rawText = rawT;
			this.rawText += "\r\n\r\n";
			cr = de =  pe = re = -1;
			try
			{
				String tempCR, tempDE, tempPE, tempRE;
				this.stockSymbol = this.rawText.substring(0, this.rawText.indexOf(" "));
				this.date = this.parseDatum("date");
				tempCR = this.parseDatum("cr");
				this.industry = this.parseDatum("industry");
				tempDE = this.parseDatum("de");
				tempPE = this.parseDatum("pe");
				tempRE = this.parseDatum("re");
				this.error = this.parseDatum("error");
				this.cr = Double.parseDouble(tempCR);
				this.de = Double.parseDouble(tempDE);
				this.pe = Double.parseDouble(tempPE);
				this.re = Double.parseDouble(tempRE);

				this.isValid = true;
			}
			catch (Exception ex)
			{
				ex.toString();
				this.isValid = false;
			}
		}
		
		/**
		 * Parses a piece of data
		 * 
		 * @param property the property to search for
		 * @return the parsed piece of string data
		 */
		private String parseDatum(String property)
		{
			int delimitLength = property.length() + 2;
			try
			{
				int startIndex;
				String temp;
				startIndex = this.rawText.lastIndexOf(this.stockSymbol + " " + property);
				if (startIndex < 0)
					return "";
				this.rawText = this.rawText.substring(startIndex);
				temp = this.rawText.substring(this.stockSymbol.length() + delimitLength, this.rawText.indexOf("\r\n"));
				return temp;
			}
			catch (Exception ex) { return ""; }
		}
	}
}