// Evan Steinberger
// CISC 466
// Spring 2007

import java.util.ArrayList;
import java.io.*;
import java.lang.reflect.*;
import java.util.Calendar;
import java.util.Date;
import java.text.*;

/**
 * The InformedAnalyzer object retrieves information from both the BlindAnalyzer and the GrabberAnalyzer, 
 * as well as other threshold information that is provided by MADSUM.  This final analysis tool is responsible 
 * for creating and coordinating all of the lower level objects and reporting the final ratios back to MADSUM.  
 * It uses a series of steps to arrive at conclusions based on yet another user defined profile.  It also uses 
 * reflection to create each of the Grabbers, so new Grabbers can be added without existing code being recompiled.
 * 
 * @author Evan Steinberger
 */

public class InformedAnalyzer
{
	private GrabberAnalyzer grabberAnal;
	private BlindAnalyzer blindAnal;
	private ArrayList grabbers;
	private String stockSymbol;

	// Profiles
	private String blindProfile;
	private String grabberProfile;
	private String informedProfile;

	// Numeric Data
	public double cr, pe, de, re;
	//private double cr_value, pe_value, de_value, re_value;

	private StockInfoGrabber.AccessThreshold accessThreshold;

	// For MADSUM specific dont care thresholds
	private double dontCareValue;
	private boolean dontCareGreater;
	private boolean useDontCare;
	private String dontCareWhich;
	
	// Profile variables
	private double BLIND_CONFIDENCE_THRESHOLD = 80;
	private double BLIND_POSITIVE_THRESHOLD = 95;
	private double GRABBER_CONFIDENCE_THRESHOLD = 70;
	private boolean USE_BEST_GRABBER = true;
	private StockInfoGrabber.AccessThreshold OLD_DATA_AGE_THRESHOLD = StockInfoGrabber.AccessThreshold.Day;

	private String historicalData;
	private double historicalValue = -1;
 
	/**
	 * Constructs an InformedAnalyzer
	 * 
	 * @param symbol the stock ticker symbol
	 * @param blindProfile which blindProfile to use
	 * @param grabberProfile which grabberProfile to use
	 * @param informedProfile which informedProfile to use
	 * @param at the Access Threshold
	 * @param useDontCare whether or not a don't care value is being use
	 * @param dontCareValue the actual don't care value
	 * @param dontCareGreater whether or not the don't care value is greater/lesser than the threshold
	 * @param dontCareWhich for which ratio the don't care value applies
	 */
	public InformedAnalyzer(String symbol, String blindProfile, String grabberProfile, String informedProfile,
		StockInfoGrabber.AccessThreshold at, boolean useDontCare, double dontCareValue, boolean dontCareGreater, String dontCareWhich)
	{
		this.stockSymbol = symbol;
		this.blindProfile = blindProfile;
		this.grabberProfile = grabberProfile;
		this.informedProfile = informedProfile;
		this.accessThreshold = at;
		this.dontCareValue = dontCareValue;
		this.dontCareGreater = dontCareGreater;
		this.useDontCare = useDontCare;
		this.dontCareWhich = dontCareWhich;

		if (!this.loadProfile(this.informedProfile))
		{
			BLIND_CONFIDENCE_THRESHOLD = .8;
			BLIND_POSITIVE_THRESHOLD = .95;
			GRABBER_CONFIDENCE_THRESHOLD = .7;
			USE_BEST_GRABBER = true;
			OLD_DATA_AGE_THRESHOLD = StockInfoGrabber.AccessThreshold.Day;
		}
		
		cr = pe = de = re = -1;
	}
	
	/**
	 * Performs all analysis.
	 */
	public void analyze()
	{
		if (!this.constructGrabbers())
		{
			System.out.println("Grabber Construction FAILED, remaining analyzation cannot continue.");
			return;
		}
		this.grabberMineInformation();
		this.performBlindAnalysis();
		this.performGrabberAnalysis();

		System.out.println("Starting Informed Caluclations ... ");
		this.cr = this.calculateValue("cr", this.grabberAnal.crBest, this.blindAnal.cr_mean, this.blindAnal.cr_confidence, this.grabberAnal.crAvg);
		this.pe = this.calculateValue("pe", this.grabberAnal.peBest, this.blindAnal.pe_mean, this.blindAnal.pe_confidence, this.grabberAnal.peAvg);
		this.de = this.calculateValue("de", this.grabberAnal.deBest, this.blindAnal.de_mean, this.blindAnal.de_confidence, this.grabberAnal.deAvg);
		this.re = this.calculateValue("re", this.grabberAnal.reBest, this.blindAnal.re_mean, this.blindAnal.re_confidence, this.grabberAnal.reAvg);
		System.out.println("Informed Calculations Complete");

		this.writeHistoricalData();
	}
	
	/**
	 * Performs an automated data collection an analysis for top NYSE stocks,
	 * based on the file topnyse.txt
	 */
	public void automatedDataCollection()
	{
		String rawStockList = ReadData.fromFile("topnyse.txt");
		String stocks[] = rawStockList.split("\r\n");
		for (String s : stocks)
		{
			this.stockSymbol = s;
			this.analyze();
		}
	}
	
	/**
	 * Constructs all StockInfoGrabbers available.
	 * 
	 * @return true if successful, false if not
	 */
	private boolean constructGrabbers()
	{
		System.out.println("Grabber Construction started ... ");
		this.grabbers = new ArrayList();
		try
		{
			File workingDir = new File(".");
			File[] files = workingDir.listFiles();
			for (File f : files)
			{
				if (f.getName().endsWith("Grabber.class") && !f.getName().equals("StockInfoGrabber.class"))
				{
					Class c = Class.forName(f.getName().substring(0, f.getName().indexOf(".class")));
					Constructor[] cons = c.getConstructors();
					Object[] args = new Object[] { this.stockSymbol, this.accessThreshold };
					Object o = cons[0].newInstance(args);
					this.grabbers.add(o);
				}
			}
			System.out.println("Grabber Construction complete");
			return true;
		}
		catch (Exception ex) { System.out.println(ex.toString()); return false; }
		
	}

	/**
	 * Instructs StockInfoGrabbers to mine all information.
	 */
	private void grabberMineInformation()
	{
		System.out.print("Parsing Grabber Data started ... ");
		for (Object o : this.grabbers)
		{
			StockInfoGrabber sig = (StockInfoGrabber)o;
			sig.mineAllInformation();
			sig.writeHistoricalContent();
		}
		System.out.println("done");
	}

	/**
	 * Instructs BlindAnalyzer to perform analysis.
	 */
	private void performBlindAnalysis()
	{
		this.blindAnal = new BlindAnalyzer(this.blindProfile);
		for (Object o : this.grabbers)
			this.blindAnal.add((StockInfoGrabber)o);
		this.blindAnal.analyze();
	}

	/**
	 * Instructs GrabberAnalyzer to perform analysis.
	 */
	private void performGrabberAnalysis()
	{
		this.grabberAnal = new GrabberAnalyzer(this.grabbers, this.blindAnal.orig, this.grabberProfile);
		this.grabberAnal.analyze();
	}
	
	/**
	 * Calculates a final return value.
	 * 
	 * @param which which ratio is being examined
	 * @param best the best Grabber to use for this ratio
	 * @param mean the mean for this ratio
	 * @param con the confidence for this ratio
	 * @param grabber_con the grabber confidence for this ratio
	 * @return the calculated return value
	 */
	private double calculateValue(String which, String best, double mean, double con, double grabber_con)
	{
		// If the confidence levels pass both thresholds defined in the profile, we have a reliable piece of information
		if ((con > BLIND_CONFIDENCE_THRESHOLD) && (grabber_con > GRABBER_CONFIDENCE_THRESHOLD))
			return mean;
		else
		{
			// If MADSUM provides a dont care threshold value, we check for it.
			if (this.useDontCare && which.equals(this.dontCareWhich))
			{
				if (this.dontCareGreater)
				{
					if (mean >= dontCareValue)
						return mean;
				}
				else
				{
					if (mean <= dontCareValue)
						return mean;
				}
			}
			
			// If the blind confidence level is high enough to ignore grabber confidence
			if (con >= BLIND_POSITIVE_THRESHOLD)
				return mean;
			
			// If profile says we should use the best grabber regardless
			
			if (USE_BEST_GRABBER)
			{
				double val = this.bestGrabberRetrieval(which, best);
				if (val != -1)
					return val;
			}

			// If arent instructed to use the best grabber (or best grabber falls through because of a bad value,
			// we can try to pull from a history if it exists and isnt too old.
			if (this.readHistoricalContent(which))
				return this.historicalValue;
			
			// Here we may end up pulling from last grabber anyway if nothing from history matches critera.
			// If it provides a bad value, -1 is the fail.
			return this.bestGrabberRetrieval(which, best);

		}
	}
	
	/**
	 * Get the value of the best Grabber for a ratio
	 * 
	 * @param which the ratio being examined
	 * @param best which Grabber is best
	 * @return the value of the best Grabber for the ratio
	 */
	private double bestGrabberRetrieval(String which, String best)
	{
		for (Object o : this.grabbers)
		{
			StockInfoGrabber sig = (StockInfoGrabber)o;
			if (sig.getClass().getName().equals(best))
				return sig.getValue(which);
		}
		return -1;
	}
	
	/**
	 * Writes data to a history file.
	 * 
	 * @return true if successful, false if not
	 */
	private boolean writeHistoricalData()
	{
		Calendar now = Calendar.getInstance();
		String orig = ReadData.fromFile("InformedResults.txt");
		if (orig.equals("!!ERROR"))
			orig = "";
		else
			orig += "\r\n\r\n";

		orig += this.stockSymbol + " date " + (now.get(Calendar.MONTH) + 1) + "/" + now.get(Calendar.DAY_OF_MONTH) + "/" + now.get(Calendar.YEAR) + " " + now.get(Calendar.HOUR_OF_DAY) + ":" + now.get(Calendar.MINUTE) + "\r\n";
		orig += this.stockSymbol + " cr " + this.cr + "\r\n";
		orig += this.stockSymbol + " pe " + this.pe + "\r\n";
		orig += this.stockSymbol + " de " + this.de + "\r\n";
		orig += this.stockSymbol + " re " + this.re + "\r\n";

		return WriteData.toFile("InformedResults.txt", orig);
		
	}
	
	/**
	 * Reads data from a history file
	 * 
	 * @param which which ratio to be examined
	 * @return true if successful, false if not
	 */
	private boolean readHistoricalContent(String which)
	{
		this.historicalData = ReadData.fromFile("InformedResults.txt");
		if (this.historicalData.equals("!!ERROR"))	
		{
			this.historicalData = "";
			return false;
		}
		if ( this.historicalData.lastIndexOf(this.stockSymbol + " date") < 0 )
			return false;

		String tempDate, tempVal;
		int startIndex = this.historicalData.lastIndexOf(this.stockSymbol + " date");
		this.historicalData = this.historicalData.substring(startIndex);
		tempDate = this.historicalData.substring(this.stockSymbol.length() + 6, this.historicalData.indexOf("\r\n"));

		// logic here to see if we should continue
		Date rightNow = new java.util.Date();
		SimpleDateFormat df = new SimpleDateFormat("MM/dd/yyyy HH:mm");
		Date previous;
		try { previous = df.parse(tempDate); }
		catch (Exception ex) { return false; }
		long diff = rightNow.getTime() - previous.getTime();
		boolean b = false;
		switch (this.OLD_DATA_AGE_THRESHOLD)
		{
			case Hour:
				if (diff >= StockInfoGrabber.HOUR)
					b = true;
				break;
			case SixHour:
				if (diff >= StockInfoGrabber.SIXHOUR)
					b = true;
				break;
			case TwelveHour:
				if (diff >= StockInfoGrabber.TWELVEHOUR)
					b = true;
				break;
			case Day:
				if (diff >= StockInfoGrabber.DAY)
					b = true;
				break;
			case Week:
				if (diff >= StockInfoGrabber.WEEK)
					b = true;
				break;
			case Month:
				if (diff >= StockInfoGrabber.MONTH)
					b = true;
				break;
			case TenMin:
				if (diff >= StockInfoGrabber.TENMIN)
					b = true;
				break;
			case Zero:
				b = true;
				break;
		}
		if (b)
			return false;

		startIndex = this.historicalData.lastIndexOf(this.stockSymbol + " " + which);
		this.historicalData = this.historicalData.substring(startIndex);
		tempVal = this.historicalData.substring(this.stockSymbol.length() + 2 + which.length(), this.historicalData.indexOf("\r\n"));

		try { this.historicalValue = Double.parseDouble(tempVal); }
		catch (Exception ex) { return false; }
		return true;
	}
	
	/**
	 * Loads an informedProfile into InformedAnalyzer
	 * 
	 * @param profileName the profile to use
	 * @return true if successful, false if not
	 */
	private boolean loadProfile(String profileName)
	{
		String content = ReadData.fromFile(profileName + ".informedprofile");
		if (content.equals("!!ERROR"))
			return false;

		try
		{
			String bct, bpt, gct, ubg, odat;
			int startIndex;

			startIndex = content.indexOf("BLIND_CONFIDENCE_THRESHOLD");
			content = content.substring(startIndex);
			bct = content.substring(27, content.indexOf("\r\n"));

			startIndex = content.indexOf("BLIND_POSITIVE_THRESHOLD");
			content = content.substring(startIndex);
			bpt = content.substring(25, content.indexOf("\r\n"));

			startIndex = content.indexOf("GRABBER_CONFIDENCE_THRESHOLD");
			content = content.substring(startIndex);
			gct = content.substring(29, content.indexOf("\r\n"));

			startIndex = content.indexOf("USE_BEST_GRABBER");
			content = content.substring(startIndex);
			ubg = content.substring(17, content.indexOf("\r\n"));

			startIndex = content.indexOf("OLD_DATA_AGE_THRESHOLD");
			content = content.substring(startIndex);
			odat = content.substring(23, content.indexOf("\r\n"));

			BLIND_CONFIDENCE_THRESHOLD = Double.parseDouble(bct);
			BLIND_POSITIVE_THRESHOLD = Double.parseDouble(bpt);
			GRABBER_CONFIDENCE_THRESHOLD = Double.parseDouble(gct);
			
			if (ubg.equals("true"))
				USE_BEST_GRABBER = true;
			else
				USE_BEST_GRABBER = false;

			OLD_DATA_AGE_THRESHOLD = StockInfoGrabber.parseAccessThreshold(odat);
			System.out.println("Loaded Informed profile: " + profileName);
			return true;
		}
		catch (Exception ex)
		{
			System.out.println("Grabber Profile " + profileName + " failed to load: " + ex.toString());
			return false;
		}

	}
	
	/**
	 * Display stock information for a grabber
	 * 
	 * @param sg the grabber to use
	 * @return a string of formatted data
	 */
	private String displayInfo(StockInfoGrabber sg)
	{
		String s = "Stock: " + sg.getStockSymbol() + " \r\n";
		s += "Industry: " + sg.getIndustry() + "\r\n";
		s += "Current Ratio: " + sg.getCurrentRatio() + "\r\n";
		s += "P/E TTM: " + sg.getPeTTM() + "\r\n";
		s += "Total DE: " + sg.getDeTotal() + "\r\n";
		s += "RE TTM: " + sg.getReTTM() + "\r\n";
		return s;
	}
	
	/**
	 * Display stock information for all grabbers.
	 */
	public String displayAllInfo()
	{
		String result="";
		for (Object o : this.grabbers)
		{
			StockInfoGrabber sig = (StockInfoGrabber)o;
			result+= "\r\n\r\n" + sig.getClass().getName() + " Provides:\r\n";
			if (sig.mineErrorMsg=="")
				result+= this.displayInfo(sig);
			else
				result+= sig.mineErrorMsg;
		}

		result+= "\r\nBlind Averages and Confidence:\r\n";
		result+= "CR: " + this.blindAnal.cr_mean + " with " + this.blindAnal.cr_confidence + "% confidence." + "\r\n";
		result+= "PE: " + this.blindAnal.pe_mean + " with " + this.blindAnal.pe_confidence + "% confidence." + "\r\n";
		result+= "DE: " + this.blindAnal.de_mean + " with " + this.blindAnal.de_confidence + "% confidence." + "\r\n";
		result+= "RE: " + this.blindAnal.re_mean + " with " + this.blindAnal.re_confidence + "% confidence." + "\r\n";

		result+= "\r\nBlind Standard Deviations:\r\n";
		result+= "CR: " + this.blindAnal.cr_std + "\r\n";
		result+= "PE: " + this.blindAnal.pe_std + "\r\n";
		result+= "DE: " + this.blindAnal.de_std + "\r\n";
		result+= "RE: " + this.blindAnal.re_std + "\r\n";

		result+= "\r\nBlind Sources Used:\r\n";
		result+= "CR: " + this.blindAnal.cr_count + "/" + this.grabbers.size() + "\r\n";
		result+= "PE: " + this.blindAnal.pe_count + "/" + this.grabbers.size() + "\r\n";
		result+= "DE: " + this.blindAnal.de_count + "/" + this.grabbers.size() + "\r\n";
		result+= "RE: " + this.blindAnal.re_count + "/" + this.grabbers.size() + "\r\n";

		result+= "\r\nBlind Probable Outliers:\r\n";
		result+= "CR: " + this.blindAnal.cr_outliers.size() + "\r\n";
		result+= "PE: " + this.blindAnal.pe_outliers.size() + "\r\n";
		result+= "DE: " + this.blindAnal.de_outliers.size() + "\r\n";
		result+= "RE: " + this.blindAnal.re_outliers.size() + "\r\n";

		result += "\r\nGrabber Analysis Results:\r\n";
		result += "CR best: " + this.grabberAnal.crBest + "\r\n";
		result += "CR worst: " + this.grabberAnal.crWorst + "\r\n";
		result += "CR Avg Confidence: " + this.grabberAnal.crAvg + "\r\n";
		result += "PE best: " + this.grabberAnal.peBest + "\r\n";
		result += "PE worst: " + this.grabberAnal.peWorst + "\r\n";
		result += "PE Avg Confidence: " + this.grabberAnal.peAvg + "\r\n";
		result += "DE best: " + this.grabberAnal.deBest + "\r\n";
		result += "DE worst: " + this.grabberAnal.deWorst + "\r\n";
		result += "DE Avg Confidence: " + this.grabberAnal.deAvg + "\r\n";
		result += "RE best: " + this.grabberAnal.reBest + "\r\n";
		result += "RE worst: " + this.grabberAnal.reWorst + "\r\n";
		result += "RE Avg Confidence: " + this.grabberAnal.reAvg + "\r\n";

		result += "\r\nFinal Results:\r\n";
		result += "CR: " + this.cr + "\r\n";
		result += "PE: " + this.pe + "\r\n";
		result += "DE: " + this.de + "\r\n";
		result += "RE: " + this.re + "\r\n";

		return result;
	}

	/*
	public static void main(String args[])
	{
		InformedAnalyzer IA = new InformedAnalyzer(args[0], "standard", "standard", "standard", StockInfoGrabber.AccessThreshold.Day, false, 0, true);
		IA.analyze();
		//IA.automatedDataCollection();
	}*/
}