import java.util.*;
import java.io.*;
/**
 * RSTTree.java
 *
 * The data structure that holds a text plan tree. See Rhetorical
 * Structure Theory, Mann & Thompson.
 *
 * An RSTTree consists of a root (an RSTNode), a parent (an RSTNode),
 * and a bid (a Bid). The bid contains information about the
 * characteristics of the tree consisting of this node and the nodes
 * below (see Bid).
 *
 * Created: Sat Dec 13 16:17:47 2003
 *
 * @author <a href="mailto:terry@Terry-Harveys-Computer.local.">Terry Harvey</a>
 * @version 1.0
 */
public class RSTTree {

    RSTNode root;
    RSTNode parent;
    //HashSet componentTrees = null;
    Bid bid = new Bid();

    static final boolean DEBUG = false;
    static String em = "Error in RSTTree: ";
    static String wm = "Warning in RSTTree: ";
    static String loc = "in RSTTree: ";

    public RSTTree() {
	root = null;
	parent = null;
    } // RSTTree constructor

    public RSTTree( RSTNode r ) {
	root = r;
	r.setTree(this);
	parent = null;
    } // RSTTree constructor

    public RSTTree(Bid b, RSTNode r){
	root = r;
	r.setTree(this);
	parent = null;
	bid = b;
    }
    
    public RSTTree(Bid b, RSTNode r, String dsORv){
	root = r;
	r.setTree(this);
	parent = null;
	bid = b;
	bid.put("topic", dsORv); //topic is string r,v,g: risk, value, or goal
    }
    
    public RSTTree(RSTNode r, String dsORv){
	root = r;
	r.setTree(this);
	parent = null;
	bid = new Bid();
	bid.put("topic", dsORv); //topic is string r,v,g: risk, value, or goal
    }
    
    public void setParent(RSTNode p){
	parent = p;
	//if (DEBUG) System.out.println(this + " setting parent " + p);
    }
    public RSTNode getParent(){
	return parent;
    }

    public void setTopic(String t){updateBid("topic", t);}
    public String getTopic(){if (bid != null) return (String)bid.get("topic"); return null;}
    public RSTNode getRoot(){ return root; }
    public Strength getTotalStrength(){return (Strength)bid.getDStot();}
    //    public RuleValue getRuleValue() {return rv;}
    public Bid getBid() {return bid;}
    public void setBid(Bid b) {bid = b;}

    public void updateBid(String attribute, Object value){
	bid.put(attribute, value);
    }

    public boolean isMixedDirection(){
	Strength total = getTotalStrength();
	if (total.direction() == 0)
	    return true;
	else return false;
    }

    public boolean isMixedTopic(){
	if (DEBUG) System.out.println("bid is: " + bid);
	String topic = getTopic();
	if (topic.equals("mixed")) return true;
	String dsType = "DS" + topic + "_exp";
	//	if (getTotalStrength() == null)
	if (DEBUG) System.out.println("dsType is: " + dsType);
	
	Strength sTopic = (Strength)bid.get(dsType);

	if (sTopic.equals(getTotalStrength()))
	    return false;
	else return true;
    }

    //returns one bid representing the combination of the bids in these trees
    public static Bid getOneBidFromTrees(Vector v, Agent a){
	BidSet bs = new BidSet(a.getName());
	Iterator i = v.iterator();
	while (i.hasNext())
	    bs.add( ((RSTTree)i.next()).getBid() );
	return bs.makeOneBidFromBidSet(a.getName());
    }

    //When merging trees, Utility must take into account sums of DS as well as max
    public void updateUtility(BidUtility bu){
	bid.getUtility(bu);
    }

    static boolean isUnifiable(RSTTree t1, RSTTree t2){
	if (DEBUG) System.out.println(loc + " isUnifiable: \n" + t1.getString() + "\n" + t2.getString());
	
	return RSTNode.isUnifiable(t1.getRoot(), t2.getRoot());
    }
    static boolean isUnifiable(RSTTree t1, RSTTree t2, Hashtable h){
	return RSTNode.isUnifiable(t1.getRoot(), t2.getRoot(), h);
    }

    /**
     * Pass in a proposition and a strength; act is default "buy investment".
     Example: 
     RSTTree t3 = GoalDo("Short term debt is very low");
     (GOAL10 (DO7 BuyInvestment ) )
         (RECOMMEND9 BuyInvestment )
         (PERSUADE8 (DO7 BuyInvestment ) )
             (MOTIVATE6 BuyInvestment Short term debt is very low )
                 (INFORM5 Short term debt is very low )
     **/
    public static RSTTree GoalDo(String prop, String RVorG, double s, Bid bid){
	RSTNode i = new RSTNode("INFORM", prop);
	RSTTree iTree = new RSTTree(i, RVorG);
	String act = "BuyInvestment";
	RSTNode m = new RSTNode("MOTIVATE", act, prop);
	m.addChild(iTree);
	Predicate d = new Predicate("DO", act);
	RSTNode p = new RSTNode("PERSUADE", d);
	p.addChild(new RSTTree(m, RVorG));
	RSTNode r = new RSTNode("RECOMMEND", act);
	RSTNode g = new RSTNode("GOAL", d);
	g.addChild(new RSTTree(r, RVorG));
	g.addChild(new RSTTree(p, RVorG));
	RSTTree t = new RSTTree(bid, g);
	t.setTopic(RVorG);
	String dsType = "DS" + RVorG + "_exp";
	t.updateBid(dsType, new Strength(s));
	iTree.setBid((Bid)t.getBid().clone());
	if (DEBUG) System.out.println(loc + "new GoalDo1: ");
	if (DEBUG) t.print();
	return t;
    }
    //get Strength from existing Bid.
    public static RSTTree GoalDo(String prop, String RVorG, Bid bid){
	RSTNode i = new RSTNode("INFORM", prop);
	RSTTree iTree = new RSTTree(i, RVorG);
	String act = "BuyInvestment";
	RSTNode m = new RSTNode("MOTIVATE", act, prop);
	m.addChild(iTree);
	Predicate d = new Predicate("DO", act);
	RSTNode p = new RSTNode("PERSUADE", d);
	p.addChild(new RSTTree(m, RVorG));
	RSTNode r = new RSTNode("RECOMMEND", act);
	RSTNode g = new RSTNode("GOAL", d);
	g.addChild(new RSTTree(r, RVorG));
	g.addChild(new RSTTree(p, RVorG));
	RSTTree t = new RSTTree(bid, g);
	t.setTopic(RVorG);
	iTree.setBid((Bid)t.getBid().clone());
	if (DEBUG) System.out.println(loc + "new GoalDo2: ");
	if (DEBUG) t.print();
	
	return t;
    }

    public static RSTTree GoalDoNot(String prop, String RVorG, double s, Bid bid){
	RSTNode i = new RSTNode("INFORM", prop);
	RSTTree iTree = new RSTTree(i, RVorG);
	String act = "BuyInvestment";
	RSTNode m = new RSTNode("DEMOTIVATEACT", act, prop);
	m.addChild(iTree);
	Predicate d = new Predicate("DO", act);
	RSTNode p = new RSTNode("DISSUADE", d);
	p.addChild(new RSTTree(m, RVorG));
	RSTNode r = new RSTNode("RECOMMEND", new Predicate("NOT", d));
	RSTNode g = new RSTNode("GOAL", new Predicate("NOT", d));
	g.addChild(new RSTTree(r, RVorG));
	g.addChild(new RSTTree(p, RVorG));
	RSTTree t = new RSTTree(bid, g);
	t.setTopic(RVorG);
	String dsType = "DS" + RVorG + "_exp";
	t.updateBid(dsType, new Strength(s));
	iTree.setBid((Bid)t.getBid().clone());
	if (DEBUG) System.out.println(loc + "new GoalDoNot1: ");
	if (DEBUG) t.print();
	return t;
    }
    public static RSTTree GoalDoNot(String prop, String RVorG, Bid bid){
	RSTNode i = new RSTNode("INFORM", prop);
	RSTTree iTree = new RSTTree(i, RVorG);
	String act = "BuyInvestment";
	RSTNode m = new RSTNode("DEMOTIVATEACT", act, prop);
	m.addChild(iTree);
	Predicate d = new Predicate("DO", act);
	RSTNode p = new RSTNode("DISSUADE", d);
	p.addChild(new RSTTree(m, RVorG));
	RSTNode r = new RSTNode("RECOMMEND", new Predicate("NOT", d));
	RSTNode g = new RSTNode("GOAL", new Predicate("NOT", d));
	g.addChild(new RSTTree(r, RVorG));
	g.addChild(new RSTTree(p, RVorG));
	RSTTree t = new RSTTree(bid, g);
	t.setTopic(RVorG);
	iTree.setBid((Bid)t.getBid().clone());
	if (DEBUG) System.out.println(loc + "new GoalDoNot2: ");
	if (DEBUG) t.print();
	
	return t;
    }

    //get Strength from existing Bid.
    public static RSTTree GoalDoAct(String act, String prop, String RVorG, Bid bid){
	RSTNode i = new RSTNode("INFORM", prop);
	RSTTree iTree = new RSTTree(i, RVorG);
	RSTNode m = new RSTNode("MOTIVATE", act, prop);
	m.addChild(iTree);
	Predicate d = new Predicate("DO", act);
	RSTNode p = new RSTNode("PERSUADE", d);
	p.addChild(new RSTTree(m, RVorG));
	RSTNode r = new RSTNode("RECOMMEND", act);
	RSTNode g = new RSTNode("GOAL", d);
	g.addChild(new RSTTree(r, RVorG));
	g.addChild(new RSTTree(p, RVorG));
	RSTTree t = new RSTTree(bid, g);
	t.setTopic(RVorG);
	iTree.setBid((Bid)t.getBid().clone());
	if (DEBUG) System.out.println(loc + "new GoalDoAct3: ");
	if (DEBUG) t.print();
	
	return t;
    }

    //get Strength from existing Bid.
    public static RSTTree GoalDoNotAct(String act, String prop, String RVorG, Bid bid){
	RSTNode i = new RSTNode("INFORM", prop);
	RSTTree iTree = new RSTTree(i, RVorG);
	RSTNode m = new RSTNode("DEMOTIVATEACT", act, prop);
	m.addChild(iTree);
	Predicate d = new Predicate("DO", act);
	RSTNode p = new RSTNode("DISSUADE", d);
	p.addChild(new RSTTree(m, RVorG));
	RSTNode r = new RSTNode("RECOMMEND", new Predicate("NOT", d));
	RSTNode g = new RSTNode("GOAL", new Predicate("NOT", d));
	g.addChild(new RSTTree(r, RVorG));
	g.addChild(new RSTTree(p, RVorG));
	RSTTree t = new RSTTree(bid, g);
	t.setTopic(RVorG);
	iTree.setBid((Bid)t.getBid().clone());
	if (DEBUG) System.out.println(loc + "new GoalDoNot3: ");
	if (DEBUG) t.print();
	
	return t;
    }
			
    //look for a node in this tree that can unify with n. Return first
    //node found. 
    public RSTNode find(RSTNode n){
	if (DEBUG) System.out.println("Entering find: " + n);
	
	return find(root, n);
    }
    public RSTNode find(RSTNode tree, RSTNode n){
	if (DEBUG) System.out.println("Entering sub find: "+ tree + " " + n);
	if (RSTNode.isUnifiable(tree, n)) return tree;
	if (DEBUG) System.out.println("sub find getting children ");
	Iterator i = tree.getChildren();
	Object o1, o2;
	while (i.hasNext()){
	    o1 = i.next();
	    if (o1 instanceof RSTTree){
		o2 = find(((RSTTree)o1).getRoot(), n); 
		if (o2 instanceof RSTNode)
		    return (RSTNode)o2;
	    }
	}
	return null;
    }

    public void fillInUtilities(BidUtility bu){
	
	if ( ! getBid().getBidder().equals("unset"))
	    updateUtility(bu);
	//System.out.println("fillInUtilities, 2nd bid: " + getBid());

	Iterator c = root.getChildren();
	while(c.hasNext()){
	    ((RSTTree)c.next()).fillInUtilities(bu);
	}
    }


    public void write (FileWriter fw) throws IOException {
	if (this.root != null){
	    StringBuffer s = new StringBuffer("\n");
	    s.append(bid.toString());
	    s.append("\n");
	    fw.write(s.toString());
	    nodeWrite( fw, "");
	}
    }

    public void write (String filename, boolean appendFlag) throws IOException {
	System.out.println("writing RSTTree to " + filename );
	FileWriter fw = new FileWriter(filename, appendFlag);
	if (this.root != null){
	    fw.write("\n");
	    nodeWrite( fw, "");
	}
	fw.flush();
	fw.close();
    }

    private void nodeWrite(FileWriter fw, String indent) throws IOException{
	RSTNode node = root;
	fw.write(indent + node.toString() + bid.utility + "\n");
	Iterator c = node.getChildren();
	while(c.hasNext()){
	    ((RSTTree)c.next()).nodeWrite( fw, indent + "   ");
	}
    }

    public void writeLong (FileWriter fw) throws IOException {
	fw.write("\n");
	if (this.root != null){
	    nodeWriteLong( fw, "");
	}
    }

    private void nodeWriteLong(FileWriter fw, String indent) throws IOException{
	StringBuffer s = new StringBuffer();
	if (getTopic() != null && 
	    (getTotalStrength().absCurr() > 0)){
	    s.append(bid.toString());
	    s.append("\n");
	}
	s.append(indent + root.toString() + "\n");
	fw.write(s.toString());
	Iterator c = root.getChildren();
	while(c.hasNext()){
	    ((RSTTree)c.next()).nodeWriteLong( fw, indent + "   ");
	}
    }

    public void writeInforms (FileWriter fw) throws IOException {
	if (this.root != null){
	    nodeWriteInforms( fw, "");
	}
    }

    private void nodeWriteInforms(FileWriter fw, String indent) throws IOException{
	StringBuffer s = new StringBuffer(100);
	if ( root.getName().regionMatches(true, 0, "INFORM", 0, 6)){
	    //s.append(bid.toString());
	    //s.append("\n");
	    s.append("\n" + indent + root.toString());
	    
	    s.append(" " + getBid().getBidder() + " " +  bid.utility  + "\n");
	    fw.write(s.toString());
	}
	Iterator c = root.getChildren();
	while(c.hasNext()){
	    ((RSTTree)c.next()).nodeWriteInforms( fw, indent + "   ");
	}
    }

	
    public void print(){
	subPrint("");
    }

    private void subPrint(String indent){
	RSTNode n = root;
	System.out.println(indent +  bid);//"s=" + str + " rv=" + rv + " " +
	System.out.println(indent + n);
        Iterator i = n.getChildren();
	while (i.hasNext()){
	    Object o = i.next();
	    if (o instanceof RSTTree)
		((RSTTree)o).subPrint(indent + "    ");
	    else if (o instanceof String)
		System.out.println(indent + o);
	    else
		System.out.println(em + " subPrint " + o);
	}
    }

    /** Sets DS to zero throughout tree for testing purposes **/
    public void zeroDS(String t){
	Strength zeroDS = new Strength( 0 );
	subZeroDS(t, zeroDS);
    }

    private void subZeroDS(String t, Strength zeroDS){

	RSTNode n = root;
	updateBid("DS" + t + "_exp", zeroDS);

        Iterator i = n.getChildren();
	while (i.hasNext()){
	    Object o = i.next();
	    if (o instanceof RSTTree)
		((RSTTree)o).subZeroDS(t, zeroDS);
	}
    }

    public String getString(){
	StringBuffer s = new StringBuffer("");
	subGetString(s, true);
	//s.append("]");
	//remove parens that confuse DECAF msg parser
	//	String st = (s.toString()).replace('(', '[');
	//	st = st.replace(')', ']');
	return s.toString();
    }

    private void subGetString(StringBuffer s, boolean startFlag){
	boolean flag = false;
	RSTNode n = root;
	s.append("[");
	s.append(bid.toString());
	s.append(n.getString());
        Iterator i = n.getChildren();
	s.append("[");
	flag = true;
	    //}
	while (i.hasNext()){
	    Object o = i.next();
	    if (o instanceof RSTTree)
		((RSTTree)o).subGetString(s, false);
	    else if (o instanceof String)
		s.append(o);
	    else
		s.append(em + " in subGetString " + o + "\n");
	}
	s.append("]"); //end child list
	s.append("]");
    }

    /*
     * Reads a tree in from the text produced by getString(). Assumes
     * that all underscores other than in node names should be spaces.
     */
    public static RSTTree readRSTTree(String s){

	StringTokenizer st = new StringTokenizer(s," []{},=", true );
	if (! st.nextToken().equals("["))
	    System.out.println(em + "readRSTTree did not find starting [ before tree in " + s);
	
	return readRSTTree(st);
    }//    public static RSTTree readRSTTree(String s){

    public static Vector readMultipleRSTTrees(String s){
	Vector trees = new Vector(6);
	String token = "dummy";
	StringTokenizer st = new StringTokenizer(s," []{},=", true );
	while(st.hasMoreTokens()){
	    token = st.nextToken();
	    if (token.equals("empty"))
		{}
	    else if ( token.equals("[") ){
		RSTTree r = readRSTTree(st);
		if (DEBUG) System.out.println(loc + "Parsed tree being added: ");
		if (DEBUG) r.print();
		trees.addElement(r);
	    }
	    //if (token.equals(" ")) continue;
	    //else if (token.equals("empty"))
	}
	return trees;
    }//  static Vector readMultipleRSTTrees(String s)

    //strip Bid, Strength and RuleValue
    public static RSTTree readRSTTree(StringTokenizer st){
	while( ! st.nextToken().equals("{")); //skip bracket of Bid
	Bid b = Bid.readBid(st);
	return new RSTTree(b, readRSTTreeNode(st));
    }
    public static RSTNode readRSTTreeNode(StringTokenizer st){
	//"[str rv [name args] [list of children]]" but missing first [
	if (! st.nextToken().equals("["))
	    System.out.println(em + "readRSTTree did not find starting [ before node in sb: " + st.toString());
	RSTNode r = RSTNode.readRSTNode(st);
	readChildList(st, r);
	if (! st.nextToken().equals("]"))
	    System.out.println(em + "readRSTTree did not find final ] after node " + r);
	
	return r;
    }//     public static RSTTree readRSTTree(StringTokenizer st)

    //add children to node
    public static void readChildList(StringTokenizer st, RSTNode r){
	if (! st.nextToken().equals("["))
	    System.out.println(em + "readChildList did not find [ after node " + r);
	String temp = st.nextToken();
	if (temp.equals("]")) //list is empty
	    return;
	while (temp.equals("[")){
	    r.addChild(readRSTTree(st));
	    temp = st.nextToken();
	}
	if (! temp.equals("]"))//(! st.nextToken().equals("]"))
	    System.out.println(em + "readChildList did not find final ] after node " + r);
    }

    /** Slow. Makes a deep copy. **/
    public Object clone(){
	return readRSTTree(this.getString());
    }

    public static void main(String[] args){

	RSTTree t3 = GoalDoNot("Short term debt is very low", DSType.RISK, -3, new Bid());//strength???
	t3.print();
	System.out.println(" t3.isMixedTopic(): " + t3.isMixedTopic());
	
	System.out.println("getString() prints: " + t3.getString());
	String temp = t3.getString();
	t3 = readRSTTree(temp);
	System.out.println("t3 post read: ");
	t3.print();
	RSTTree t4 = (RSTTree)t3.clone();
	t4.print();
	System.out.println("t4 above ");
	
	t3 = readRSTTree(temp);
	System.out.println("getString() prints: " + t3.getString());
	System.out.println("t3 post read: ");
	t3.print();


	System.out.println("asdf asddf a sdf  asdf a sdf  asd f as df a sddf ass df a sdf a sd f as ddf a sdf a sd fa sdf ");
	
	
    }
} // RSTTree


