import java.io.*;
import java.util.*;
/**
 * RuleSet.java
 *
 * Created: Dec 10 20:12:01 2003
 *
 * This class contains rules for integrating RSTTrees.
 *   
 * Try to unify input trees with the template trees; if
 * unifiable, then return the RSTTree resulting from the
 * integration of the two trees. Else return null.
 *
 * It would be nice to be able to mark the rule trees with adjoin
 * points, and then mark the trees being unified at the relevant
 * portions as they are being traversed; then simply make one point
 * the child of another. Could rewrite tree unify to do the marking
 * after unifiability has been determined.
 *
 * If a tree is used whole, then I clone the existing tree object;
 * if it needs to be modified, I create a new one so that I don't
 * modify other trees which may use the object I am working on.

 * Rule1: Joins props inside INFORM nodes using AND; same direction, same strength, same topic (positive for now)
 * Rule2:  CONCESSION: one tree has DS 2 more than other, diff directions;
 * Rule3: CONTRAST requires trees to be the same (or 1 different) strength; could be diff topic
 * Rule4: concession or contrast of mixed topics
 * Rule5: ELABORATION for positive trees, same topic, with Strength that is not identical
 * Rule6: JOINT under Motivation, for trees with diff topic
 */
public class RuleSet {

    static boolean DEBUG = false;
    static String em = "Error in RuleSet.java: ";
    static String wm = "Warning in RuleSet.java: ";
    static String loc = "RuleSet.java: ";
    static Vector currentRules = null;

    public RuleSet() {
	if (currentRules == null){
	    currentRules = new Vector();
	    currentRules.add(new Rule1());
	    currentRules.add(new Rule2());
	    currentRules.add(new Rule3());
	    currentRules.add(new Rule4());
	    currentRules.add(new Rule5());
	    currentRules.add(new Rule7());
	    currentRules.add(new Rule6());
	}
	else
	    System.out.println("RuleSet using existing rules");
    } // RuleSet constructor
    
    //whichRules is an ordered vector of rule numbers to put in this set
    public RuleSet(Vector whichRules) {
	if (currentRules == null){
	    currentRules = new Vector();
	    Iterator i = whichRules.iterator();
	    Integer nextRule;
	    while(i.hasNext()){
		nextRule = (Integer)i.next();
		switch(nextRule.intValue()){
		case 1: currentRules.add(new Rule1()); break;
		case 2: currentRules.add(new Rule2()); break;
		case 3: currentRules.add(new Rule3()); break;
		case 4: currentRules.add(new Rule4()); break;
		case 5: currentRules.add(new Rule5()); break;
		case 6: currentRules.add(new Rule6()); break;
		case 7: currentRules.add(new Rule7()); break;
		default: 
		    System.out.println(em + "in constructor, rule number: " + nextRule);
		    
		}
	    }
	    System.out.println("RuleSet using custom rule set: " + currentRules);
	}
	else
	    System.out.println("RuleSet using existing rules");
    } // RuleSet constructor
    
    Iterator getRules(){return currentRules.iterator();}

    /**
     * Make new tree for a Rule by merging the components of two
     * trees.  Does not merge the tree structures themselves, only
     * combines their attributes (bids) and puts the combination in
     * the root of the new tree
     **/
    public static RSTTree makeNewTreeFromTwo(RSTTree a, RSTTree b, RSTNode newRoot, int ruleNumber, int ruleValue, String agentName){
	RSTTree	newTree = new RSTTree(newRoot);
	if (DEBUG) System.out.println("makeNewTreeFromTwo: " + ruleNumber);
	if (DEBUG) System.out.println("makeNewTreeFromTwo: " + newRoot.toString());
	if (DEBUG) System.out.println("a root" + a.getRoot());
	if (DEBUG) System.out.println("b root" + b.getRoot());
		
	//new bid will be combined from a,b, plus new rule value
	BidSet bs = new BidSet(agentName);
	bs.add(a.getBid());
	bs.add(b.getBid());
	newTree.setBid( bs.makeOneBidFromBidSet(agentName) );
	System.out.println("************rn: " + ruleNumber + " rv: " + ruleValue);

	if (DEBUG) System.out.println("makeNewTreeFromTwo: old rv: " + (RuleValue)newTree.getBid().getValue("RV"));

	//int tempBackwardsRuleValue;
	//if (ruleValue == 1) tempBackwardsRuleValue = 2;
	//else if  (ruleValue == 2) tempBackwardsRuleValue = 1;
	//else tempBackwardsRuleValue = ruleValue;

	//newTree.getBid().increaseRuleValue(ruleNumber, tempBackwardsRuleValue); //add value for this rule
	newTree.getBid().increaseRuleValue(ruleNumber, ruleValue); //add value for this rule
	if (DEBUG) System.out.println("makeNewTreeFromTwo: new rv: " + (RuleValue)newTree.getBid().getValue("RV"));
	if (a.getTopic().equals(b.getTopic())) newTree.setTopic(a.getTopic());
	else  newTree.setTopic(DSType.MIXED);

	return newTree;
    }

    public Vector applyRules(RSTTree t1, RSTTree t2, String agentName){
	return tryRulesOnNodesInTree(t1, t1, t2, agentName, true);
    }

    /*
     * Try to use rules to add entire t2 to various points in tree t;
     * return Vector of all possible new trees.
     */
    //t1CurrRoot is a subtree of t1
    private Vector tryRulesOnNodesInTree(RSTTree t1, RSTTree t1CurrRoot, RSTTree t2, String agentName, boolean firstPass){
	if (DEBUG) System.out.println("t1 CurrRoot" + t1CurrRoot.getRoot());
	if (DEBUG) System.out.println("t1 root" + t1.getRoot());
	if (DEBUG) System.out.println("t2 root" + t2.getRoot());

	Vector results = new Vector();
	RSTTree newTree, wholeNew;
	Rule nextRule;

	if (t1CurrRoot.getTopic() == null){
	    if (DEBUG) System.out.println(loc + "tryRules stopped, no topic set");
	    return results;
	}

	Iterator rules = getRules();
	while(rules.hasNext()){
	    nextRule = (Rule)rules.next();

	    newTree = null;

	    if (firstPass == true ||
		(firstPass == false && nextRule.worksOnSubTrees())){

		if (DEBUG) System.out.println("************rn1: " + nextRule.ruleNumber() + " rv1: " + nextRule.ruleValue());
		if (t1.isMixedDirection()){  
		    if (DEBUG) System.out.println("mixed!");
		
		    if (nextRule.worksUnderMixedDirection()){
			if (DEBUG) System.out.println("works under mixed direction: " + nextRule);
		    
			newTree = nextRule.apply(t1CurrRoot, t2, agentName);
		    }
		    else newTree = null;
		}
		else newTree = nextRule.apply(t1CurrRoot, t2, agentName);
	    }
	    if (newTree != null){
		wholeNew = (RSTTree)t1.clone();
		if (DEBUG) System.out.println("rule: " + nextRule);
		if (DEBUG) System.out.println("wholeNew root" + wholeNew.getRoot());
		if (DEBUG) System.out.println("t1CurrRoot root" + t1CurrRoot.getRoot());
		//find node in wholeNew matching t1CurrRoot root
		RSTNode here = wholeNew.find(t1CurrRoot.getRoot());
		if (DEBUG) System.out.println("here: " + here);
		RSTTree hereTree = here.getTree();
		if (hereTree == wholeNew){
		    results.add(makeNewTreeFromTwo(t1, t2, newTree.getRoot(),
						   nextRule.ruleNumber(), nextRule.ruleValue(), agentName));
		}
		else{
		    RSTNode parent = hereTree.getParent();
		    if (! parent.removeChild(hereTree)) //delete t1cr subtree from wholeNew
			System.out.println(em + "child removal failed for " + 
					   hereTree + " in " + parent);
		    //parent.addChild(newTree);
		    parent.addChild(makeNewTreeFromTwo(t1CurrRoot, t2, newTree.getRoot(), //not quite
						       nextRule.ruleNumber(), nextRule.ruleValue(), agentName));
		    results.add(makeNewTreeFromTwo(t1, t2, wholeNew.getRoot(), 
				  nextRule.ruleNumber(), nextRule.ruleValue(), agentName));
		}
	    }
	}//while
		
	results.addAll( tryRuleOnT1Children(t1, t1CurrRoot, t2, agentName) );
	return results;
    }//tryRulesOnNodesInTree(RSTTree t1, RSTTree t1CurrRoot, RSTTree t2, String agentName, boolean firstPass){

    private Vector tryRuleOnT1Children(RSTTree wholeT1, RSTTree t1, RSTTree t2, String agentName){
	Vector results = new Vector();
	RSTNode root = t1.getRoot();
	Iterator i = root.getChildren();
	RSTTree t3;
	while(i.hasNext()){
	    if (DEBUG) System.out.println("while");
	    if (DEBUG) t1.print();
	    t3 = (RSTTree)i.next();
	    if (DEBUG) t3.print();
	    results.addAll(tryRulesOnNodesInTree(wholeT1, t3, t2, agentName, false));
	}
	return results;
    }

    /*
    //Aggregation under contrast or concession using Elaboration
    //b is a simple tree (ruleValue==0) 
    public RSTTree Rule4(RSTTree a, RSTTree b, String agentName){
	int ruleNum = 4, ruleVal = 1;

	Strength sa = a.getTotalStrength();
	Strength sb = b.getTotalStrength();
	if (DEBUG) System.out.println("in rule4 " + sa.direction() + " " + sb.direction() + " sa:" + sa + " sb:" + sb);

	RSTNode c1 = new RSTNode("CONTRAST");
	RSTNode c2 = new RSTNode("CONCESSION");
	Hashtable h = new Hashtable(10);
	if ( ! (RSTNode.isUnifiable(c1, a.getRoot(), h) ||
		RSTNode.isUnifiable(c2, a.getRoot(), h)))
	    return null;

	if (DEBUG) System.out.println("Rule4 GOT HERE3!");

	Iterator i = a.getRoot().getChildren();
	RSTTree c3 = (RSTTree)i.next();
	RSTTree c4 = (RSTTree)i.next();

	RSTTree newTree = null;
	RSTTree attempt = null;//Rule5(c3, b, agentName);
	if (attempt != null){
	    RSTNode newC = (RSTNode)a.getRoot().clone(); may not clone correctly, check children
	    newC.addChild(attempt);
	    newC.addChild( (RSTTree)c4.clone() );
	    newTree = makeNewTreeFromTwo(a, b, newC, 4, 1, agentName);
	}
	else {
	    attempt = null;//Rule5(c4, b, agentName);
	    if (attempt == null)
		return null;
	    RSTNode newC = (RSTNode)a.getRoot().clone();
	    newC.addChild( (RSTTree)c3.clone() );
	    newC.addChild(attempt);
	    newTree = makeNewTreeFromTwo(a, b, newC, 4, 1, agentName);
	}
	    
	if (attempt == null)
	    return null;

	return newTree;
    }// Rule4(RSTTree a, RSTTree b) aggregation under concession or contrast
*/


    public static void main(String[] args){
	Bid b1 = new Bid();
	b1.put("length_exp", "10");
	Bid b2 = new Bid();
	b2.put("length_exp", "9");
	Bid b3 = new Bid();
	b3.put("length_exp", "8");
	Bid b4 = new Bid();
	b4.put("length_exp", "7");

	Vector rv = new Vector();
	rv.add(new Integer(2));
	rv.add(new Integer(3));
	rv.add(new Integer(1));
	rv.add(new Integer(5));
	rv.add(new Integer(7));
	rv.add(new Integer(6));
	rv.add(new Integer(4));

	RuleSet r = new RuleSet(rv);

	RSTTree x  = RSTTree.readRSTTree("[{topic=r, length_exp=26.0, DStot_exp=[[2.5 3.94 2][0 0 0]], cost_exp=6.0, DSr_exp=[[2.5 3.94 2][0 0 0]], DSg_exp=[[0 0 0][0 0 0]], DSv_exp=[[0 0 0][0 0 0]], RV=[[5] 1], symbol=IBM, U=1.6185259451920242, bidder=DR}[Goal [DO BuyInvestment ]][[{DSv_exp=[[0 0 0][0 0 0]], DSr_exp=[[0 0 0][0 0 0]], DSg_exp=[[0 0 0][0 0 0]], U=0.5180340182519565, bidder=unset}[Recommend BuyInvestment ][]][{DSv_exp=[[0 0 0][0 0 0]], DSr_exp=[[0 0 0][0 0 0]], DSg_exp=[[0 0 0][0 0 0]], U=0.5180340182519565, bidder=unset}[Persuade [DO BuyInvestment ] ][[{DSv_exp=[[0 0 0][0 0 0]], DSr_exp=[[0 0 0][0 0 0]], DSg_exp=[[0 0 0][0 0 0]], U=null, bidder=unset}[MOTIVATE BuyInvestment [Elaborate ] ][[{DSv_exp=[[0 0 0][0 0 0]], DSr_exp=[[0 0 0][0 0 0]], DSg_exp=[[0 0 0][0 0 0]], U=null, bidder=unset}[Elaborate ][[{topic=r, length_exp=15.0, DStot_exp=[[2.5 2.5 1][0 0 0]], cost_exp=2, DSr_exp=[[2.5 2.5 1][0 0 0]], DSg_exp=[[0 0 0][0 0 0]], DSv_exp=[[0 0 0][0 0 0]], symbol=IBM, RV=[[] 0], U=0.9980720958509005, bidder=DE}[Inform company_has_a_low_debt_equity_ratio_comma_suggesting_the_ability_to_weather_an_economic_downturn][]][{topic=r, length_exp=11.0, DStot_exp=[[1.44 1.44 1][0 0 0]], cost_exp=4, DSr_exp=[[1.44 1.44 1][0 0 0]], DSg_exp=[[0 0 0][0 0 0]], DSv_exp=[[0 0 0][0 0 0]], symbol=IBM, RV=[[] 0], U=0.8330756204547319, bidder=CR}[Inform company_has_a_strong_current_ratio_relative_to_the_industry_average][]]]]]]]]]]");
	RSTTree y = RSTTree.readRSTTree("[{topic=r, length_exp=7.0, DStot_exp=[[0 0 0][0.5 0.5 1]], cost_exp=2, DSr_exp=[[0 0 0][0.5 0.5 1]], DSg_exp=[[0 0 0][0 0 0]], DSv_exp=[[0 0 0][0 0 0]], symbol=IBM, RV=[[] 0], U=0.7531862779895032, bidder=IR}[GOAL [NOT [DO BuyInvestment ] ] ][[{DSv_exp=[[0 0 0][0 0 0]], topic=r, DSr_exp=[[0 0 0][0 0 0]], DStot_exp=[[0 0 0][0 0 0]], DSg_exp=[[0 0 0][0 0 0]], U=0.5180340182519565, bidder=unset}[RECOMMEND [NOT [DO BuyInvestment ] ] ][]][{DSv_exp=[[0 0 0][0 0 0]], topic=r, DSr_exp=[[0 0 0][0 0 0]], DStot_exp=[[0 0 0][0 0 0]], DSg_exp=[[0 0 0][0 0 0]], U=0.5180340182519565, bidder=unset}[DISSUADE [DO BuyInvestment ] ][[{DSv_exp=[[0 0 0][0 0 0]], topic=r, DSr_exp=[[0 0 0][0 0 0]], DStot_exp=[[0 0 0][0 0 0]], DSg_exp=[[0 0 0][0 0 0]], U=null, bidder=unset}[DEMOTIVATEACT BuyInvestment industry_has_a_somewhat_aggressive_risk_profile;][[{topic=r, length_exp=7.0, DStot_exp=[[0 0 0][0.5 0.5 1]], cost_exp=2, DSr_exp=[[0 0 0][0.5 0.5 1]], DSg_exp=[[0 0 0][0 0 0]], DSv_exp=[[0 0 0][0 0 0]], symbol=IBM, RV=[[] 0], U=0.7531862779895032, bidder=IR}[INFORM industry_has_a_somewhat_aggressive_risk_profile;][]]]]]]]]");
	RSTTree w = RSTTree.GoalDoNot("goaldo2", DSType.RISK, -2, b3);
	//RSTTree w2 = RSTTree.GoalDoNot("goaldo3", DSType.RISK, -3, b4);

	Vector trees = new Vector();
	trees.add(x);
	trees.add(y);
	//trees.add(w);
	//trees.add(w2);
	BidUtility bu = BidUtility.readBidUtilityFromString("{cost_coeff=0, cost_target=25, length_coeff=1, length_target=100, DSr_coeff=5, DSr_target=10, DSv_coeff=1, DSv_target=5, DSg_target=5, DSg_coeff=1, RV_coeff=1}");
	//BidUtility bu = BidUtility.readBidUtilityFromString("{cost_coeff=1, cost_target=25, length_coeff=1, length_target=60, DSr_coeff=5, DSr_target=10, DSv_coeff=10, DSv_target=10, DSg_target=10, DSg_coeff=5, RV_coeff=1}");
	RSTTreeComp rc = new RSTTreeComp(bu);
	
	
	Collections.sort(trees,rc);
	Integrator treeIntegrator = new Integrator(bu);
	
	//trees = treeIntegrator.addNewTrees(trees, "spam");
	   
	for(int i = 0; i < trees.size(); i++)
	    ((RSTTree)trees.get(i)).updateUtility(bu);
	System.out.println("first all trees: ");
	for(int i = 0; i < trees.size(); i++)
	    ((RSTTree)trees.get(i)).print();
	
	//Vector v = r.tryRulesOnNodesInTree(x, x, y, "spam");
	//System.out.println("vector:");
	//for(int i = 0; i < v.size(); i++)
	//  ((RSTTree)v.elementAt(i)).print();
	
	//System.out.println("vector end");
	//RSTTree z = (RSTTree)v.elementAt(0);//null;//r.Rule3(x, y,  "spam");
	////RSTTree z2 = r.Rule3(x, y, "spam");
	//
	//v = r.tryRulesOnNodesInTree(z, z, w, "spam");
	//System.out.println("vector:");
	//for(int i = 0; i < v.size(); i++)
	//  ((RSTTree)v.elementAt(i)).print();
	
	trees.insertElementAt(w, 0);
	Collections.sort(trees,rc);
	RSTTree last = treeIntegrator.joinAllTrees(trees, "spuk");

	last.print();

	//or(int i = 0; i < trees.size(); i++)
	//   ((RSTTree)trees.get(i)).updateUtility(bu);
	//   
	//Collections.sort(trees,rc);
	//RSTTree z = (RSTTree)trees.elementAt(0);//null;//r.Rule3(x, y,  "spam");
	//
	//StringBuffer s = new StringBuffer("\n");
	//TreeGen.gen(s, z, bu);
	//System.out.println(s.toString());
	//
	////w.print();
	////x.print();
	////y.print();
	////
        //if (z!= null){// && z2 != null
	//    System.out.println("ZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZ");
	//    try {
	//	System.out.println("writing RSTTree to file trees" );
	//	FileWriter fw = new FileWriter("trees", true);
	//	fw.write("______________________________________________________________\n");
	//	//w.write(fw);
	//	//x.write(fw);
	//	//y.write(fw);
	//	z.write(fw);
	//	//z2.write(fw);
	//	//fw.write("\n" + z.getString() + "\n" );
	//	fw.flush();
	//	fw.close();
	//    } 
	//    catch (Exception e) {
	//	System.out.println(e);
	//    }
	//    z.print();
        //}

    }

} // RuleSet
