// Evan Steinberger
// CISC 466
// Spring 2007

import java.text.*;
import java.util.Date;
import java.util.Calendar;

/**
 * StockInfoGrabber is the base class from which all
 * other specific grabber classes are derived.
 * 
 * @author Evan Steinberger
 */

public abstract class StockInfoGrabber
{
	public static final long TENMIN = 1000 * 60 * 10;
	public static final long HOUR = 1000 * 60 * 60;
	public static final long SIXHOUR = HOUR * 6;
	public static final long TWELVEHOUR = SIXHOUR * 2;
	public static final long DAY = TWELVEHOUR * 2;
	public static final long WEEK = DAY * 7;
	public static final long MONTH = DAY * 30;

	private boolean readNewData = false;
	
	protected String stockSymbol;
	protected String pageHTML;
	protected boolean validSymbol;
	public String mineErrorMsg;

	protected double currentRatio, peTTM, deTotal, reTTM;
	protected String industry;

	private String historicalDate;

	
	protected AccessThreshold currentAccessThreshold = AccessThreshold.Hour;
	public static enum AccessThreshold { Zero, TenMin, Hour, SixHour, TwelveHour, Day, Week, Month };

	/**
	 * Parses a string that contains an Access Threshold
	 * 
	 * @param s	the String to be parsed
	 * @return	an Access Threshold
	 * @exception	Bad AccessThreshold
	 */
	public static AccessThreshold parseAccessThreshold(String s) throws Exception
	{
		if (s.toLowerCase().equals("month"))
			return AccessThreshold.Month;
		else if (s.toLowerCase().equals("week"))
			return AccessThreshold.Week;
		else if (s.toLowerCase().equals("day"))
			return AccessThreshold.Day;
		else if (s.toLowerCase().equals("twelvehour"))
			return AccessThreshold.Day;
		else if (s.toLowerCase().equals("sixhour"))
			return AccessThreshold.SixHour;
		else if (s.toLowerCase().equals("hour"))
			return AccessThreshold.Hour;
		else if (s.toLowerCase().equals("zero"))
			return AccessThreshold.Zero;
		else if (s.toLowerCase().equals("tenmin"))
			return AccessThreshold.TenMin;
		else
			throw new Exception("Bad AccessThreshold");
	}


	// Whether or not the InfoGrabber will always go to the web to get info
	// or try to look at its own history first
	protected boolean alwaysDownload = false;
	// Whether or not to always record historical data
	protected boolean alwaysRecord = true;
	protected String historicalData;
	protected String histCopy;
	// set this initially to false
	protected boolean hasHistoricalContent = false;
	
	/**
	 * Sets the StockInfoGrabber to always download data from the web
	 * 
	 * @param b	boolean value to set
	 */
	public void setAlwaysDownload(boolean b)
	{
		this.alwaysDownload = b;
	}

	/**
	 * Sets the StockInfoGrabber to always record historical data
	 * 
	 * @param b	boolean value to set
	 */
	public void setAlwaysRecord(boolean b)
	{
		this.alwaysRecord = b;
	}

	/**
	 * Sets the StockInfoGrabber's current Access Threshold
	 * 
	 * @param at the AccessThreshold
	 */
	public void setThreshold(AccessThreshold at)
	{
		this.currentAccessThreshold = at;
	}

	/**
	 * Gets one of four important ratios from the StockInfoGrabber
	 * 
	 * @param which the ratio to get, in string form of "cr","pe","de", or "re"
	 * @return	the numeric equivalent to the ratio, or -1 for error
	 */
	public double getValue(String which)
	{
		which = which.toLowerCase();
		if (which.equals("cr"))
			return currentRatio;
		else if (which.equals("pe"))
			return peTTM;
		else if (which.equals("de"))
			return deTotal;
		else if (which.equals("re"))
			return reTTM;
		else
			return -1;
	}
	
	/**
	 * Performs General initialization for StockInfoGrabber
	 * 
	 * @param stockSymbol	the actual stock ticker symbol
	 */
	protected void init(String stockSymbol)
	{
		this.stockSymbol = stockSymbol.toUpperCase();
		this.initValues();
		this.historicalData="";
		this.historicalDate="";
	}

	/**
	 * Performs value initialization for StockInfoGrabber
	 * 
	 */
	private void initValues()
	{
		this.currentRatio = peTTM = deTotal = reTTM = -1;
		this.pageHTML = "";
		this.mineErrorMsg = "";
		this.industry = "Unknown";
	}
	
	/**
	 * Downloads HTML from a URL
	 * 
	 * @param url	the address from which to get data
	 */
	protected void downloadHTML(String url)
	{
		if (this.hasHistoricalContent)
			return;

		if (this.alwaysDownload)
			this.pageHTML = WebQuery.getCompleteHTML(url);
		else
		{
			this.hasHistoricalContent = this.readHistoricalContent();
			if (!this.hasHistoricalContent)
			{
				this.readNewData = true;
				this.pageHTML = WebQuery.getCompleteHTML(url);
			}
			else
			{
				Date rightNow = new java.util.Date();
				SimpleDateFormat df = new SimpleDateFormat("MM/dd/yyyy HH:mm");
				Date previous;
				try
				{
					previous = df.parse(this.historicalDate);
				}
				catch (Exception ex)
				{
					this.readNewData = true;
					this.pageHTML = WebQuery.getCompleteHTML(url);
					return;
				}
				long diff = rightNow.getTime() - previous.getTime();
				boolean b = false;
				switch (this.currentAccessThreshold)
				{
					case Hour:
						if (diff >= HOUR)
							b = true;
						break;
					case SixHour:
						if (diff >= SIXHOUR)
							b = true;
						break;
					case TwelveHour:
						if (diff >= TWELVEHOUR)
							b = true;
						break;
					case Day:
						if (diff >= DAY)
							b = true;
						break;
					case Week:
						if (diff >= WEEK)
							b = true;
						break;
					case Month:
						if (diff >= MONTH)
							b = true;
						break;
					case TenMin:
						if (diff >= TENMIN)
							b = true;
						break;
					case Zero:
						b = true;
						break;
				}
				if (b)
				{
					this.readNewData = true;
					this.pageHTML = WebQuery.getCompleteHTML(url);
				}
				else
					return;
			}
		}
	}
	
	/**
	 * Validates the Stock Symbol is valid.
	 * Implemented by specific grabbers.
	 */
	protected abstract void validateSymbol();

	/**
	 * Builds a URL for a specific stock symbol and Grabber
	 * Implemented by specific grabbers.
	 * 
	 * @param s	any additional information in the URL
	 */
	protected abstract String buildURL(String s);
	
	/**
	 * Mines CR.
	 * Implemented by specific grabbers.
	 */
	protected abstract boolean mineCurrentRatio();

	/**
	 * Mines PE.
	 * Implemented by specific grabbers.
	 */
	protected abstract boolean minePeTTM();

	/**
	 * Mines DE.
	 * Implemented by specific grabbers.
	 */
	protected abstract boolean mineDeTotal();

	/**
	 * Mines RE.
	 * Implemented by specific grabbers.
	 */
	protected abstract boolean mineReTTM();

	/**
	 * Mines Industry.
	 * Implemented by specific grabbers.
	 */
	protected abstract boolean mineIndustry();

	/**
	 * Mines all Information
	 * Implemented by specific grabbers.
	 */
	public abstract boolean mineAllInformation();
	
	/**
	 * Gets CR
	 * 
	 * @return the current ratio
	 */
	public double getCurrentRatio() { return this.currentRatio; }

	/**
	 * Gets PE
	 * 
	 * @return the price/earnings ratio
	 */
	public double getPeTTM() { return this.peTTM; }

	/**
	 * Gets DE
	 * 
	 * @return the debt/equity ratio
	 */
	public double getDeTotal() { return this.deTotal; }

	/**
	 * Gets RE
	 * 
	 * @return the return on equity
	 */
	public double getReTTM() { return this.reTTM; }
	
	/**
	 * Gets Stock Symbol
	 * 
	 * @return the current stock symbol
	 */
	public String getStockSymbol() { return this.stockSymbol; }

	/**
	 * Gets Industry
	 * 
	 * @return the Industry
	 */
	public String getIndustry() { return this.industry; }
	
	/**
	 * Write historical Data out to GrabberHistoryFile
	 * 
	 * @return true if successful, false if error
	 */
	public boolean writeHistoricalContent()
	{
		if (!alwaysRecord)
			return true;

		// no sense in duplicating the record
		if (!this.readNewData)
			return true;

		String orig="";
		if (this.historicalData == "")
			orig = ReadData.fromFile(this.getClass().getName() + ".txt");
		else
			orig = this.historicalData;

		if (orig.equals("!!ERROR"))
			orig="";
		else
			orig+= "\r\n\r\n";

	
		Calendar now = Calendar.getInstance();
		orig+= this.stockSymbol + " " + "date" + " " + (now.get(Calendar.MONTH) + 1) + "/" + now.get(Calendar.DAY_OF_MONTH) + "/" + now.get(Calendar.YEAR) + " " + now.get(Calendar.HOUR_OF_DAY) + ":" + now.get(Calendar.MINUTE) + "\r\n";
		orig+= this.stockSymbol + " " + "cr" + " " + currentRatio + "\r\n";
		orig+= this.stockSymbol + " " + "industry" + " " + industry + "\r\n";
		orig+= this.stockSymbol + " " + "de" + " " + deTotal + "\r\n";
		orig+= this.stockSymbol + " " + "pe" + " " + peTTM + "\r\n";
		//orig+= this.stockSymbol + " " + "type" + " " + currentRatio + "\r\n";
		orig+= this.stockSymbol + " " + "re" + " " + reTTM + "\r\n";
		orig+= this.stockSymbol + " " + "error ";
		if (this.mineErrorMsg != "")
			orig+= this.mineErrorMsg;
		else
			orig+= "none";
		orig += "\r\n";
		return WriteData.toFile(this.getClass().getName() + ".txt", orig);
	}

	/**
	 * Reads historical Data from GrabberHistoryFile
	 * 
	 * @return true if successful, false if error
	 */
	protected boolean readHistoricalContent()
	{
		this.historicalData = ReadData.fromFile(this.getClass().getName() + ".txt");
		if (this.historicalData.equals("!!ERROR"))	
		{
			this.historicalData = "";
			return false;
		}
		if ( this.historicalData.lastIndexOf(this.stockSymbol + " date") < 0 )
			return false;

		String tempDate, tempCR, tempPE, tempDE, tempRE, tempIndustry, tempError;
		this.histCopy = new String();
		this.histCopy = this.historicalData.toString();
		
		// Must be in order to work right and for max efficiency
		tempDate = this.parseDatum(6, "date");
		tempCR = this.parseDatum(4, "cr");
		tempIndustry = this.parseDatum(10, "industry");
		tempDE = this.parseDatum(4, "de");
		tempPE = this.parseDatum(4, "pe");
		tempRE = this.parseDatum(4, "re");
		tempError = this.parseDatum(7, "error");

		// if there was an error in the last entry the data is probably trash.
		// might implement some other logic here to go back to a previous entry
		// in the historical data but for now it will just return false
		if (!tempError.equals("none") || tempDate.equals("") ||
			tempCR.equals("") || tempIndustry.equals("") || tempDE.equals("") ||
			tempPE.equals("") || tempRE.equals(""))
			return false;

		try
		{
			this.currentRatio = Double.parseDouble(tempCR);
			this.industry = tempIndustry;
			this.deTotal = Double.parseDouble(tempDE);
			this.peTTM = Double.parseDouble(tempPE);
			this.reTTM = Double.parseDouble(tempRE);
			this.historicalDate = tempDate;
		}
		catch (Exception ex) { return false; }
		return true;
	}
	
	/**
	 * Parses a particular piece of historic data
	 * 
	 * @param delimitLength	length of delimiter for historic data
	 * @param property	the property to be parsed
	 * @return a string of parsed data
	 */
	private String parseDatum(int delimitLength, String property)
	{
		try
		{
			int startIndex;
			String temp;
			startIndex = this.histCopy.lastIndexOf(this.stockSymbol + " " + property);
			if (startIndex < 0)
				return "";
			this.histCopy = this.histCopy.substring(startIndex);
			temp = this.histCopy.substring(this.stockSymbol.length() + delimitLength, this.histCopy.indexOf("\r\n"));
			return temp;
		}
		catch (Exception ex) { return ""; }

	}

}